# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

import cupy as cp

def _disk_picking(umin: float, umax: float, n: int, seed: int = None, radial_uniformity: bool = False, dtype: type = cp.float32) -> cp.ndarray:
    """Generate `n` points uniformly distributed within a disk.

    This function generates points within a disk in the x-y plane, centered at (0, 0, 0).
    The distribution of points can be either uniform across the disk's area or uniform across
    the radius.

    Parameters
    ----------
    umin : float
        The minimum normalized radius of the disk.
    umax : float
        The maximum normalized radius of the disk.
    n : int
        The number of points to generate.
    seed : int, optional
        Seed for the random number generator. If None, a random seed will be used.
    radial_uniformity : bool, optional
        If False, the points are distributed uniformly across the disk's area.
        If True, the points are distributed uniformly along the radius.
        Defaults to True.
    dtype : type, optional
        The desired data type of the returned array. Defaults to `cp.float32`.

    Returns
    -------
    ps : cp.ndarray, shape (n, 3)
        A CuPy array containing the generated points. Each row represents a point (x, y, z),
        where z is always 0.

    Notes
    -----
    - If `radial_uniformity` is False, the radial distance of the points from the center is proportional to the square root of a uniform random variable.
    This ensures a uniform distribution of points across the disk's area.
    - If `radial_uniformity` is True, the radial distance is directly proportional to the uniform random variable.
    - The angular coordinate of the points is uniformly distributed between 0 and 2*pi.
    """
    rng = cp.random.default_rng(cp.random.Philox4x3210(seed))

    u = rng.random(n, dtype=dtype) 
    if not radial_uniformity:
        u = cp.sqrt(u*(umax**2 - umin**2) + umin**2) 
    else:
        u = u*(umax - umin) + umin

    v = rng.random(n, dtype=dtype) * 2. * cp.pi

    ps = cp.empty((u.shape[0], 3), dtype=dtype)
    ps[:, 0] = u * cp.cos(v)
    ps[:, 1] = u * cp.sin(v)
    ps[:, 2] = 0.
    return ps



# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

import time
from typing import (
    Tuple, List, 
    Union, Optional
)

import cupy as cp
import numpy as np
from tqdm.auto import tqdm
import random

from .optics._optical_system import OpticalSystem
from .optics.utils._aspheric_coeffs import normalize_aspheric_coefficients
from .optics._cpu_transforms import local_to_telescope_rotation
from .optics._surface_misc import SurfaceShape, SensorSurfaceType
from .optics._atmosphere import AtmosphericTransmission
from .optics._materials import Materials
from .optics import ray_tracing

from .electronics import CherenkovCamera, CherenkovSiPMCamera

from .utils._timer import BenchmarkTimer


class IACT:
    """A class to represent an Imaging Atmospheric Cherenkov Telescope (IACT).

    This class serves as a wrapper for a CUDA-accelerated backend. It handles the conversion 
    of optical system and camera information into CuPy ndarrays (transferring data to the GPU), 
    and arranges this data in a format suitable for input to the CUDA kernels that perform the 
    actual simulation.

    Parameters
    ----------
    optical_system : OpticalSystem
        An instance of a class derived from :py:class:`OpticalSystem` representing the 
        telescope optical system.
    camera : CherenkovCamera
        An instance of a class derived from :py:class:`CherenkovCamera` representing the 
        telescope Cherenkov camera.
    position : Tuple[float, float, float] | List[float] | numpy.ndarray
        The Cartesian coordinates (x, y, z) representing the telescope 
        position in millimeters. Can be a tuple, list, or NumPy array.
    pointing : Tuple[float, float] | List[float] | numpy.ndarray
        The horizontal coordinates (altitude, azimuth) representing the 
        telescope pointing direction in degrees. Can be a tuple, list, 
        or NumPy array.    
    
    Raises
    ------
    TypeError
        
        - If optical_system is not an instance of :py:class:`OpticalSystem` or a derived class.
        - If camera is not an instance of :py:class:`CherenkovCamera` or a derived class.

    ValueError
        
        - If position does not have 3 elements.
        - If pointing does not have 2 elements.

    """

    def __init__(
        self,
        optical_system: OpticalSystem,
        camera: CherenkovCamera = None,
        position: Tuple[float, float, float] | List[float] | np.ndarray = (0.,0.,0.),
        pointing: Tuple[float, float] | List[float] | np.ndarray = (0,0),
    ):
        if not isinstance(optical_system, OpticalSystem):
            raise TypeError("optical_system must be an instance of OpticalSystem or a derived class")

        # Validate and convert position to NumPy array
        if not isinstance(position, (tuple, list, np.ndarray)) or len(position) != 3:
            raise ValueError("position must be a tuple, list, or NumPy array of length 3")

        self._cuda_tracing_args = [None] * 5 # ray-tracing kernel related arguments
        self._cuda_atmosphere_args = [None] * 5 # atmosphere transmission related arguments
        self._cuda_materials_args = [None] * 5 # materials refractive index args
        
        self._position = np.asarray(position)

        self.pointing = pointing

        self._optical_system = optical_system
        self._camera = camera

        self.atmospheric_transmission = AtmosphericTransmission()

        # Benchmark
        self.timer = BenchmarkTimer()
        self.timer.add_section('simulate_response')

        self.show_progress = False

        self._blocksize = 128 # Threads per block

    @property
    def position(self) -> np.ndarray:
        """
        numpy.ndarray: The telescope position in Cartesian coordinates (x, y, z) 
                    in millimeters.
        """
        return self._position

    @position.setter
    def position(self, value: Tuple[float, float, float] | List[float] | np.ndarray):
        if not isinstance(value, (tuple, list, np.ndarray)) or len(value) != 3:
            raise ValueError("position must be a tuple, list, or NumPy array of length 3")
        self._position = np.asarray(value)
        self._cuda_telescope_init()

    @property
    def pointing(self) -> np.ndarray:
        """
        numpy.ndarray: The telescope pointing direction in horizontal 
                    coordinates (altitude, azimuth) in degrees.
        """
        return self._pointing

    @pointing.setter
    def pointing(self, value: Tuple[float, float] | List[float] | np.ndarray):
        if not isinstance(value, (tuple, list, np.ndarray)) or len(value) != 2:
            raise ValueError("pointing must be a tuple, list, or NumPy array of length 2")
        self._pointing = np.asarray(value)
        self._altitude = value[0]
        self._azimuth = value[1]
        self._cuda_telescope_init()
    
    @property
    def altitude(self) -> float:
        """
        float: Telescope pointing altitude (0-90).
        """
        return self._altitude
    
    @altitude.setter
    def altitude(self, value: float):
        if value < 0 or value > 90:
            raise(ValueError('Telescope pointing altitude must be between 0 (horizon) and 90 (zenith) degrees.'))
        self._altitude = value
        self._pointing[0] = value
        self._cuda_telescope_init()

    @property
    def azimuth(self) -> float:
        """
        float: Telescope pointing azimuth (0-360).
        """
        return self._azimuth
    
    @azimuth.setter
    def azimuth(self, value: float):
        if value < 0 or value > 360:
            raise(ValueError('Telescope pointing azimuth must be between 0 and 360 degrees (increasing from north to east).'))
        self._azimuth = value
        self._pointing[1] = value
        self._cuda_telescope_init()
    
    @property
    def optical_system(self) -> OpticalSystem:
        """
        OpticalSystem: The telescope optical system
        """
        return self._optical_system
    
    @property
    def camera(self) -> CherenkovCamera:
        """
        CherenkovCamera: The telescope Cherenkov camera.
        """
        return self._camera

    @camera.setter
    def camera(self, a_camera):
        if a_camera is not None and not isinstance(a_camera, CherenkovCamera):
            raise TypeError("camera must be an instance of CherenkovCamera or a derived class")
        self._camera = a_camera

    def _cuda_telescope_init(self):
        # Position and rotation to transform into the telescope reference frame
        telescope_rot = cp.asarray(local_to_telescope_rotation(*self.pointing), dtype=cp.float32)
        telescope_pos = cp.asarray(self.position, dtype=cp.float32)
        
        self._cuda_tracing_args[0] = telescope_pos
        self._cuda_tracing_args[1] = telescope_rot
    
    def _cuda_surfaces_init(self):
        """Prepares optical surface data for CUDA-based ray tracing.

        This function extracts relevant properties from the optical surfaces
        defined in :py:attr:`optical_system` and converts them into CuPy arrays.
        """

        NUM_ASPHERIC_COEFFS = 10
        ROT_MATRIX_SIZE = 9

        surface_data_dtype = np.dtype([
            ('position',              np.float32, 3),                   # 12 bytes
            ('_pad1',                 np.uint8,   4),                   # ! Padding to align 'offset' (float2) to 8 bytes (12+4=16)
            ('offset',                np.float32, 2),                   # 8 bytes (starts at 16)
            ('rotation_matrix',       np.float32, ROT_MATRIX_SIZE),     # 36 bytes (starts at 24)
            ('curvature',             np.float32),                      # 4 bytes (starts at 60)
            ('conic_constant',        np.float32),                      # 4 bytes (starts at 64)
            ('aspheric_coefficients', np.float32, NUM_ASPHERIC_COEFFS), # 40 bytes (starts at 68, ends at 108)
            ('_pad2',                 np.uint8,   4),                   # ! Padding to align 'aperture_size' to 112
            ('aperture_size',         np.float32, 2),                   # 8 bytes (starts at 112)
            ('aperture_shape',        np.int32,   2),                   # 8 bytes
            ('flags',                 np.bool_,   2),                   # 2 bytes
            # align=True will automatically handle the 2-byte padding needed here for 'type' (int)
            ('type',                  np.int32),                        # 4 bytes
            ('shape',                 np.int32),                        # 4 bytes
            ('material_before',       np.int32),                        # 4 bytes
            ('material_after',        np.int32),                        # 4 bytes
            ('scattering_dispersion', np.float32),                      # 4 bytes
            ('sensor_info',           np.float32, 6),                   # 24 bytes
        ], align=True)

        n_surfaces = len(self.optical_system)

        h_surfaces = np.zeros(n_surfaces, dtype=surface_data_dtype)

        sensor_counter = 0
        for i, surface in enumerate(self.optical_system):
            if surface._shape not in SurfaceShape:
                raise(ValueError(f"Surface shape '{surface._shape}' ray-tracing not yet implemented."))

            h_surfaces[i]['position'] = np.asarray(surface.position, dtype=np.float32)
            h_surfaces[i]['rotation_matrix'] = np.asarray(surface.get_rotation_matrix().flatten(), dtype=np.float32)
            h_surfaces[i]['type'] = np.int32(surface.type.value)
            h_surfaces[i]['shape'] = np.int32(surface._shape.value)
            h_surfaces[i]['material_before'] = np.int32(surface.material_in.value)
            h_surfaces[i]['material_after'] = np.int32(surface.material_out.value)
            h_surfaces[i]['scattering_dispersion'] = np.float32(surface.scattering_dispersion/180.)

            if surface._shape == SurfaceShape.CYLINDRICAL:
                h_surfaces[i]['aperture_size'][0] = np.float32(surface.radius)
                h_surfaces[i]['aperture_size'][1] = np.float32(surface.height)
                h_surfaces[i]['flags'][0] = np.bool_(surface.top)
                h_surfaces[i]['flags'][1] = np.bool_(surface.bottom)
                continue

            h_surfaces[i]['aperture_size'][0] = np.float32(surface.half_aperture)
            h_surfaces[i]['aperture_size'][1] = np.float32(surface.central_hole_half_aperture)
            h_surfaces[i]['aperture_shape'][0] = np.int32(surface.aperture_shape.value)
            h_surfaces[i]['aperture_shape'][1] = np.int32(surface.central_hole_shape.value)

            # Register pixels (in the order they have been inserted in the optical system)
            sensor_type = surface.sensor_type
            h_surfaces[i]['sensor_info'][0] = np.float32(sensor_type.value)
            if surface.sensor_type is not SensorSurfaceType.NONE:
                if sensor_type == SensorSurfaceType.SIPM_TILE:
                    h_surfaces[i]['sensor_info'][0] = np.float32(surface.sensor_type.value)
                    h_surfaces[i]['sensor_info'][1] = np.float32(sensor_counter)
                    h_surfaces[i]['sensor_info'][2] = np.float32(surface.pixels_per_side)
                    h_surfaces[i]['sensor_info'][3] = np.float32(surface.pixel_active_side)
                    h_surfaces[i]['sensor_info'][4] = np.float32(surface.pixels_separation)
                    h_surfaces[i]['sensor_info'][5] = np.float32(surface.border_to_active_area)
                sensor_counter += surface._n_pixels

            if surface._shape == SurfaceShape.FLAT:
                continue
            
            h_surfaces[i]['offset'] = np.asarray(surface.offset, dtype=np.float32)
            h_surfaces[i]['curvature'] = np.float32(surface.curvature)
            h_surfaces[i]['flags'][0] = np.bool_(surface.is_fresnel)

            if surface._shape == SurfaceShape.ASPHERICAL:
                h_surfaces[i]['conic_constant'] = np.float32(surface.conic_constant)
                surface_asph_coeffs = normalize_aspheric_coefficients(surface.aspheric_coefficients, surface.half_aperture)
                for j in range(len(surface_asph_coeffs)):
                    h_surfaces[i]['aspheric_coefficients'][j] = np.float32(surface_asph_coeffs[j])

        
        self._cuda_tracing_args[2] = cp.asarray(h_surfaces.view(np.uint8), dtype=cp.uint8)

    def _cuda_surfaces_efficiency_init(self):
        optical_tables_dtype = np.dtype([
            ('transmittance_front', np.uint64),
            ('reflectance_front',   np.uint64),
            ('transmittance_back',  np.uint64),
            ('reflectance_back',    np.uint64),
            ('efficiency',          np.uint64),
            ('opt_inv_dwl',         np.float32),
            ('opt_start_wl',        np.float32),
            ('opt_inv_dang',        np.float32),
            ('opt_start_ang',       np.float32),
            ('eff_inv_dwl',         np.float32),
            ('eff_start_wl',        np.float32),
            ('eff_inv_dang',        np.float32),
            ('eff_start_ang',       np.float32)
        ], align=True)
        
        n_surfaces = len(self.optical_system)
        h_optical_tables = np.zeros(n_surfaces, dtype=optical_tables_dtype)
        self._surface_textures = [] # keep a reference to the textures to avoid garbage collection
        for k,s in enumerate(self.optical_system):
            surface_prop = s.properties
            if surface_prop is None or not surface_prop.is_defined:
                continue
            opt_textures = surface_prop.get_textures()
            self._surface_textures.append(opt_textures)
            h_optical_tables[k]['transmittance_front'] = opt_textures.transmittance.ptr
            h_optical_tables[k]['reflectance_front'] = opt_textures.reflectance.ptr
            h_optical_tables[k]['transmittance_back'] = opt_textures.transmittance_back.ptr
            h_optical_tables[k]['reflectance_back'] = opt_textures.reflectance_back.ptr
            h_optical_tables[k]['opt_inv_dwl'] = opt_textures.optical_wavelength_inv_step
            h_optical_tables[k]['opt_start_wl'] = opt_textures.optical_wavelength_start
            h_optical_tables[k]['opt_inv_dang'] = opt_textures.optical_angle_inv_step
            h_optical_tables[k]['opt_start_ang'] = opt_textures.optical_angle_start

            h_optical_tables[k]['efficiency'] = opt_textures.efficiency.ptr
            h_optical_tables[k]['eff_inv_dwl'] = opt_textures.efficiency_wavelength_inv_step
            h_optical_tables[k]['eff_start_wl'] = opt_textures.efficiency_wavelength_start
            h_optical_tables[k]['eff_inv_dang'] = opt_textures.efficiency_angle_inv_step
            h_optical_tables[k]['eff_start_ang'] = opt_textures.efficiency_angle_start
        
        self._cuda_tracing_args[3] = cp.asarray(h_optical_tables.view(np.uint8), dtype=cp.uint8)

    def _cuda_materials_init(self):
        """Prepares materials data for CUDA-based ray tracing.
        This function extracts relevant properties from all defined materials.
        """
        # Refractive indices
        refractive_tables_dtype = np.dtype([
            ('texture',  np.uint64),
            ('inv_dwl',  np.float32),
            ('start_wl', np.float32)
        ], align=True)

        n_materials = sum(1 for _ in Materials._members())

        h_refractive_tables = np.zeros(n_materials, dtype=refractive_tables_dtype)

        self._ri_textures = [] # keep a reference to the textures to avoid garbage collection
        for material in Materials._members():
            material_texture = material.get_refractive_index_texture()
            self._ri_textures.append(material_texture)
            h_refractive_tables[material.value]['texture'] = material_texture.texture.ptr
            h_refractive_tables[material.value]['inv_dwl'] = material_texture.wavelength_inv_step
            h_refractive_tables[material.value]['start_wl'] = material_texture.wavelength_start

        self._cuda_tracing_args[4] = cp.asarray(h_refractive_tables.view(np.uint8), dtype=cp.uint8)

    def _cuda_atmosphere_init(self):
        if self.atmospheric_transmission.value is not None:
            wl = self.atmospheric_transmission.wavelength
            zem = self.atmospheric_transmission.altitude
            self._cuda_atmosphere_args[:] = [
                cp.asarray(self.atmospheric_transmission.value, dtype=cp.float32),
                cp.asarray(wl, dtype=cp.float32),
                cp.asarray(zem, dtype=cp.float32),
                cp.asarray([len(wl), len(zem)], dtype=cp.int32),
            ]
                
    def cuda_init(self):
        """Build and copy to device all the CUDA kernels input related to the optical configuration."""

        self._cuda_surfaces_init()

        self._cuda_materials_init()

        self._cuda_telescope_init()

        self._cuda_surfaces_efficiency_init()

        if isinstance(self.camera, CherenkovSiPMCamera):
            pass

        self._cuda_atmosphere_init()

        self._d_num_surfaces = cp.int32(len(self.optical_system))

    def trace_photons(self,
            positions: Union[np.ndarray, cp.ndarray],
            directions: Union[np.ndarray, cp.ndarray],
            wavelengths: Union[np.ndarray, cp.ndarray],
            arrival_times: Union[np.ndarray, cp.ndarray],
            emission_altitudes: Optional[Union[np.ndarray, cp.ndarray]] = None,
            events_mapping: Union[np.ndarray, list] = None,
            *,
            photons_per_bunch: int = 1,
            transform_to_tel = True,
            simulate_camera = False,
            min_n_pes = 1,
            get_camera_input: bool = False,
            max_bounces=100,
            save_last_bounce=False,
            reset_state=True,
            seed: Optional[int] = None,
        ):
        """Traces the paths of photons through the telescope optical system and optionally simulates the Cherenkov camera response.

        This function simulates the propagation of photons through an optical system,
        taking into account their initial positions, directions, wavelengths, and arrival times.
        It can optionally:
        
          - model the effects of atmospheric transmission based on emission altitudes;
          - simulate the Cherenkov camera response.

        Each element in the input arrays can represent either a single photon or a "bunch" of photons.
        The ``photons_per_bunch`` parameter determines how many photons are represented by each element.
        If ``photons_per_bunch`` > 1, it is assumed that all photons within a bunch share the same
        initial position, direction, wavelength, and arrival time.

        Parameters
        ----------
        positions : Union[numpy.ndarray, cp.ndarray] of shape (N, 3)
            Initial Cartesian coordinates (x, y, z) of the photon bunches in millimeters (mm).
            Must be either a NumPy (CPU) or CuPy (GPU) array.
        directions : Union[numpy.ndarray, cp.ndarray] of shape (N, 3)
            Initial direction vectors (vx, vy, vz) of the photon bunches.
            These should be normalized. Must be either a NumPy (CPU) or CuPy (GPU) array.
        wavelengths : Union[numpy.ndarray, cp.ndarray] of shape (N,)
            Wavelengths of the photon bunches in nanometers (nm).
            Must be either a NumPy (CPU) or CuPy (GPU) array.
        arrival_times : Union[numpy.ndarray, cp.ndarray] of shape (N,)
            Arrival times of the photon bunches at their respective positions in nanoseconds (ns).
            Must be either a NumPy (CPU) or CuPy (GPU) array.
        emission_altitudes : Optional[Union[numpy.ndarray, cp.ndarray]], optional
            Emission altitudes of the photon bunches in millimeters (mm).
            This is used to calculate atmospheric transmission effects if enabled.
            Must be either a NumPy (CPU) or CuPy (GPU) array.
        events_mapping : numpy.ndarray or list
            A NumPy array of shape (N+1,) to split the photons in to groups corresponding to N different events when 
            generating the camera input arrays.
            ``events_mapping[i]`` indicates the starting index in ``positions`` for the i-th event.
            ``events_mapping[i+1]`` indicates the ending index (exclusive) in ``positions`` for the i-th event.
            Therefore, photons belonging to event i are located in ``positions[events_mapping[i]:events_mapping[i+1], :]``.
            The last element of events_mapping must be equal to the number of photons.
        photons_per_bunch : int, optional
            Number of photons represented by each element in the input arrays. If 1, each element represents a single photon.
            If > 1, each element represents a bunch of photons that share the same properties, by default 1. Maximum value is 65535.
        transform_to_tel: bool, optional
            Whether to move detected photons into the telescope refrence frame or in the detector(s) reference frame. By default True.
        simulate_camera : bool, optional
            Whether to simulate the response of the telescope Cherenkov camera.
        min_n_pes : int, optional
            Minimum number of photo-electrons needed to trigger a camera simulation (if ``simulate_camera`` is True). By default 1.
        get_camera_input : bool, optional
            Wheter to return the input for camera simulation (i.e. photon arrival times for each pixel) as CuPy ndarrays.
            By default False.
        max_bounces : int, optional
            Maxium number of bounces before the photon is killed. By default 100.
        save_last_bounce : float, optional
            Keep the position at the last bounce. Only for ray tracing visualization. By default False.
        reset_state : bool, optional
            Keep the ID of the last surface reached without re-initialize the _d_weights array. Only for ray tracing visualization. By default True.
        seed : int, optional
            Base seed of the simulation. If None a random seed will be used. By default None.

        Returns
        -------
        ndarray
            List of triggered events index. If ``simulate_camera`` is True and ``get_camera_input`` is False.
        tuple(cp.ndarray, cp.ndaray)
            Input CuPy arrays for :py:meth:`CherenkovCamera.simulate_response()`` method. If ``simulate_camera`` is False and ``get_camera_input`` is True.
        tuple(cp.ndarray, cp.ndaray, np.ndarray)
            Input CuPy arrays for :py:meth:`CherenkovCamera.simulate_response()`` method and the list of triggered events index. If ``simulate_camera`` is True and ``get_camera_input`` is True.

        Warning
        -------
            The input arrays can be either CuPy or NumPy ndarray. In the former case arrays are modified in place.
            In the latter case the arrays are copied to the device and can be accessed through the following attributes:

              - self._d_directions
              - self._d_positions
              - self._d_wavelengths
              - self._d_arrival_times
              - self._d_emission_altitudes
              - self._d_weights
        
        Warning
        -------
            Telescope configuration must be copied into the device before calling this method. Use the :py:meth:`cuda_init()` method.

        Notes
        -----

          - The ray-tracing operates with positions in mm, wavelengths in nm, and times in ns. It also handles photon bunches,
            where each thread may process multiple photons if ``photons_per_bunch`` > 1. In this case is assumed 
            that each element of the input arrays represents a bunch of photons that share the same 
            initial position, direction, wavelength, and arrival time. The arrays will maintain this
            structure, with each element representing the updated properties of a photon bunch, but the bunch size may
            be reduced throught the ray-tracing due to the interacrtion with optical elements, atmosphere and pixels.

        """
        if self.timer.active:
            cp.cuda.stream.get_current_stream().synchronize()
            t0 = time.time()

        # Copy photon bunches into GPU
        self._d_directions = cp.asarray(directions, dtype=cp.float32)
        self._d_positions = cp.asarray(positions, dtype=cp.float32)
        self._d_wavelengths = cp.asarray(wavelengths, dtype=cp.float32)
        self._d_arrival_times = cp.asarray(arrival_times, dtype=cp.float32)
        if emission_altitudes is not None:
            self._d_emission_altitudes = cp.asarray(emission_altitudes, dtype=cp.float32)

        # Calculate number of photons and kernel call parameters
        n_photons = self._d_positions.shape[0]
        num_blocks = int(np.ceil(n_photons / self._blocksize))
        d_num_photons = cp.int32(n_photons)

        # Initialize photon bunches size on device
        #   - This step can be skipped for visualization purposes. 
        #     Since when save_last_bounce is True the last 
        #     surface ID is stored in the last 16 bit of 
        #     each element in _d_weights.
        if reset_state:
            self._d_weights = cp.full((n_photons,), photons_per_bunch, dtype=cp.int32)
        
        # ID of the pixel reached by each bunch
        self._d_pixid = cp.empty((n_photons,), dtype=cp.int32)

        if seed is None:
            seed = random.getrandbits(63)

        if self.timer.active:
            cp.cuda.stream.get_current_stream().synchronize()
            t1 = time.time()
            self.timer.add_entry('simulate_response', 'copy to device', t1-t0)
        
        # Optional atmospheric transmission calculation
        if self.atmospheric_transmission.value is not None and emission_altitudes is not None:
            ray_tracing.atmospheric_transmission(
                (num_blocks,), 
                (self._blocksize,),
                (
                    self._d_positions,
                    self._d_directions,
                    self._d_wavelengths,
                    self._d_arrival_times,
                    self._d_emission_altitudes,
                    *self._cuda_atmosphere_args,
                    d_num_photons,
                    cp.uint64(seed)
                )
            )

        if self.timer.active:
            cp.cuda.stream.get_current_stream().synchronize()
            t2 = time.time()
            self.timer.add_entry('simulate_response', 'atmospheric transmission', t2-t1)

        # Perform ray tracing
        ray_tracing.ray_tracing(
            (num_blocks,),
            (self._blocksize,),
            (
                self._d_positions,
                self._d_directions,
                self._d_wavelengths,
                self._d_arrival_times,
                self._d_weights,
                self._d_pixid,
                *self._cuda_tracing_args,
                self._d_num_surfaces,
                d_num_photons,
                cp.int32(max_bounces),
                cp.bool_(save_last_bounce),
                cp.bool_(transform_to_tel),
                cp.uint64(seed+1)
            ),
            shared_mem = len(self.optical_system) * 4 * 4 # Surface positions and bounding radii
        )

        if self.timer.active:
            cp.cuda.stream.get_current_stream().synchronize()
            t3 = time.time()
            self.timer.add_entry('simulate_response', 'trace onto focal plane', t3-t2)

        # Build the input for `CherenkovCamera.simulate_response()` method
        if get_camera_input or simulate_camera:

            # Split photons into events
            if events_mapping is None:
                events_mapping = cp.asarray([0, n_photons], dtype=cp.int32)
            else:
                events_mapping = cp.asarray(events_mapping, dtype=cp.int32)

            n_pixels = self.camera.n_pixels
            n_events = len(events_mapping) - 1

            # Filter for detected photons to reduce workload
            hit_mask = self._d_pixid >= 0
            
            if not hit_mask.any():
                # No photons detected
                phs = cp.array([], dtype=cp.float32)
                phs_mapping = cp.zeros((n_events, n_pixels + 1), dtype=cp.int32)
                pe_mapping = cp.zeros((n_events + 1,), dtype=cp.int32)
                active_indices = np.array([], dtype=np.int32)
                pe_mapping_host = np.zeros((n_events + 1,), dtype=np.int32)
            else:
                hit_indices_global = cp.where(hit_mask)[0]
                hit_pix = self._d_pixid[hit_mask]
                hit_time = self._d_arrival_times[hit_mask]
                hit_weight = self._d_weights[hit_mask]

                # Get event IDs of each detected photon
                hit_event = cp.searchsorted(events_mapping, hit_indices_global, side='right').astype(cp.int32) - 1

                
                # Check if we need to expand bunches (weight > 1)
                # Use cumulative sum + searchsorted to expand indices.
                if cp.any(hit_weight > 1):
                    # Create the cumulative
                    offsets = cp.cumsum(hit_weight)
                    total_pes = int(offsets[-1])
                    
                    # Create destination indices [0, 1, ... total_pes-1]
                    dest_idx = cp.arange(total_pes, dtype=cp.int32)
                    
                    # Find which original photon each destination index belongs to
                    source_idx = cp.searchsorted(offsets, dest_idx, side='right')
                    
                    # Gather the data (repeated according to the weight)
                    final_pix = hit_pix[source_idx]
                    final_event = hit_event[source_idx]
                    final_time = hit_time[source_idx]
                else:
                    final_pix = hit_pix
                    final_event = hit_event
                    final_time = hit_time

                # Create composite key to sort by event, then by pixel
                composite_key = final_event.astype(cp.int64) * n_pixels + final_pix
                sort_order = cp.argsort(composite_key)
                phs = final_time[sort_order]

                # Generate mappings
                # bincount calculates how many PEs are in each (event, pixel) bin
                flat_counts = cp.bincount(composite_key, minlength=n_events*n_pixels).astype(cp.int32)
                counts_2d = flat_counts.reshape((n_events, n_pixels))

                # phs_mapping: cumulative sum along pixels for each event
                zeros_col = cp.zeros((n_events, 1), dtype=cp.int32)
                counts_padded = cp.hstack([zeros_col, counts_2d])
                phs_mapping = cp.cumsum(counts_padded, axis=1, dtype=cp.int32)

                # pe_mapping: global offsets for slicing 'phs' by event
                event_pes = counts_2d.sum(axis=1)
                pe_mapping = cp.concatenate([cp.array([0], dtype=cp.int32), cp.cumsum(event_pes, dtype=cp.int32)])

                if self.timer.active:
                    cp.cuda.stream.get_current_stream().synchronize()
                    t4 = time.time()
                    self.timer.add_entry('simulate_response', 'generate camera input', t4-t3)

            # Simulate camera response
            if simulate_camera:
                if phs.size > 0:
                    n_pes_per_event = pe_mapping[1:] - pe_mapping[:-1]
                    active_mask = n_pes_per_event >= min_n_pes
                    active_indices_gpu = cp.where(active_mask)[0]
                    
                    # Transfer minimal loop control data to CPU
                    active_indices = active_indices_gpu.get()
                    pe_mapping_host = pe_mapping.get()
                else:
                    active_indices = []
                
                if self.timer.active:
                    cp.cuda.stream.get_current_stream().synchronize()
                    t5 = time.time()
                    self.timer.add_entry('simulate_response', 'filter events to be simulated', t5-t4)

                event_number = []

                for i in tqdm(active_indices, disable=not self.show_progress):
                    start_phs = pe_mapping_host[i]
                    end_phs = pe_mapping_host[i+1]
                    self.camera.simulate_response((phs[start_phs:end_phs], phs_mapping[i]))
                    if self.camera.triggered:
                        event_number.append(i)
                
                event_number = np.array(event_number)

                if self.timer.active:
                    cp.cuda.stream.get_current_stream().synchronize()
                    t6 = time.time()
                    self.timer.add_entry('simulate_response', 'simulate camera response', t6-t5)

                if get_camera_input:
                    return phs, phs_mapping, event_number
                else:
                    return event_number
            
            if get_camera_input:
                return phs, phs_mapping
    
    def visualize_ray_tracing(self,
            positions: Union[np.ndarray, cp.ndarray],
            directions: Union[np.ndarray, cp.ndarray],
            wavelengths: Union[np.ndarray, cp.ndarray],
            arrival_times: Union[np.ndarray, cp.ndarray],
            emission_altitudes: Optional[Union[np.ndarray, cp.ndarray]] = None,
            photons_per_bunch: int = 1,
            *,
            map_wavelength_color=True,
            get_renderer = False,
            opacity=None,
            show_hits=True,
            show_rays=True, 
            point_size=1,
            orthographic=False,
            focal_point=None,
            resolution=None,
            max_bounces=100,
        ):
        """
        Visualizes the ray tracing process by performing step-by-step propagation of photons.

        This method initializes a VTK renderer and iteratively calls `trace_photons` with 
        `max_bounces=1` to capture and visualize the path of photons between surfaces.

        Parameters
        ----------
        positions : Union[numpy.ndarray, cp.ndarray] of shape (N, 3)
            Initial Cartesian coordinates (x, y, z) of the photon bunches in millimeters (mm).
            Must be either a NumPy (CPU) or CuPy (GPU) array.
        directions : Union[numpy.ndarray, cp.ndarray] of shape (N, 3)
            Initial direction vectors (vx, vy, vz) of the photon bunches.
            These should be normalized. Must be either a NumPy (CPU) or CuPy (GPU) array.
        wavelengths : Union[numpy.ndarray, cp.ndarray] of shape (N,)
            Wavelengths of the photon bunches in nanometers (nm).
            Must be either a NumPy (CPU) or CuPy (GPU) array.
        arrival_times : Union[numpy.ndarray, cp.ndarray] of shape (N,)
            Arrival times of the photon bunches at their respective positions in nanoseconds (ns).
            Must be either a NumPy (CPU) or CuPy (GPU) array.
        emission_altitudes : Optional[Union[numpy.ndarray, cp.ndarray]], optional
            Emission altitudes of the photon bunches in millimeters (mm).
            This is used to calculate atmospheric transmission effects if enabled.
            Must be either a NumPy (CPU) or CuPy (GPU) array.
        photons_per_bunch : int, optional
            Number of physical photons represented by one simulated photon. Defaults to 1.
        get_renderer : bool, optional
            If True, returns the `VTKOpticalSystem` instance instead of starting the render loop immediately. 
            Defaults to False.
        opacity : float, optional
            Opacity of the ray lines (0.0 to 1.0). If None, it is calculated based on the number 
            of photons. Defaults to None.
        show_hits : bool, optional
            Whether to render points at the location where rays intersect surfaces. Defaults to True.
        show_rays : bool, optional
            Whether to render lines representing the photon paths. Defaults to True.
        point_size : int, optional
            Size of the points rendered at intersection hits. Defaults to 1.
        orthographic : bool, optional
            If True, start with a parallel projection. 
            If False, start with a Perspective projection. Default is False.
        focal_point : tuple, list, or str, optional
            If tuple/list: (x, y, z) coordinates to look at.
            If str: The name of the surface in self.os to look at.
            If None: Defaults to center of system.
        resolution : float, optional
            Objects mesh resolution (in mm). By default None (depends on object extent).
        
        Returns
        -------
        VTKOpticalSystem or None
            Returns the renderer instance if `get_renderer` is True, otherwise returns None 
            after the render window is closed.
        """
        from .visualization._vtk_optical_system import VTKOpticalSystem
        
        positions = cp.asarray(positions, dtype=cp.float32)
        directions = cp.asarray(directions, dtype=cp.float32)
        wavelengths = cp.asarray(wavelengths, dtype=cp.float32)
        arrival_times = cp.asarray(arrival_times, dtype=cp.float32)
        emission_altitudes = cp.asarray(emission_altitudes, dtype=cp.float32) if emission_altitudes is not None else None

        first_step = True

        renderer = VTKOpticalSystem(self, resolution=resolution)
        vis_wavelengths = None

        bounces = 0
        while True:

            ps_start = positions.get()
            
            self.trace_photons(
                positions,
                directions,
                wavelengths,
                arrival_times,
                emission_altitudes,
                photons_per_bunch=photons_per_bunch, 
                simulate_camera=False,
                max_bounces=1, 
                save_last_bounce=True,
                reset_state=first_step,
                transform_to_tel=False,
            )
            survivors = cp.isfinite(wavelengths)
            positions = positions[survivors]
            directions = directions[survivors]
            wavelengths = wavelengths[survivors]
            arrival_times = arrival_times[survivors]
            self._d_weights = self._d_weights[survivors]
            if emission_altitudes is not None:
                emission_altitudes = emission_altitudes[survivors]
            
            first_step = False
            
            if map_wavelength_color:
                vis_wavelengths = wavelengths.get()
            ps_start = ps_start[survivors.get()]
            ps_stop = positions.get()

            if np.array_equal(ps_start, ps_stop, equal_nan=True):
                break

            surface_hits = (self._d_weights >> 16) & 0xFFFF
            surface_hits = surface_hits.astype(cp.int32).get() - 1
            
            renderer.add_rays(
                ps_start, ps_stop,
                wavelengths=vis_wavelengths,
                surface_id=surface_hits,
                show_hits=show_hits,
                show_rays=show_rays,
                point_size=point_size
            )

            bounces += 1
            if bounces >= max_bounces:
                break

        renderer.set_ray_opacity(opacity)

        if get_renderer:
            return renderer
        else:
            renderer.start_render(focal_point=focal_point, orthographic=orthographic)


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np

# Tolerance for considering a value to be zero
ZERO_TOLERANCE = 1e-6 # mm^-1, 1000 m curvature radius 

def sag(radius: np.ndarray, curvature: float, conic_constant: float, 
        aspheric_coefficients: np.ndarray, axial_z: float, is_fresnel: bool) -> np.ndarray:
    r"""
    Calculates the sagitta of an aspheric surface.

    Parameters
    ----------
    radius : numpy.ndarray
        Radial distance(s) from the optical axis (can be a scalar or a NumPy array).
    curvature : float
        Surface curvature (1/radius of curvature).
    conic_constant : float
        Conic constant.
    aspheric_coefficients : numpy.ndarray
        1D array of aspheric coefficients. The i-th element is the coefficient
        of the r**(2*(i+1)) term.
    axial_z : float
        Axial z-coordinate of the surface vertex.
    is_fresnel : bool
        Whether the surface is a Fresnel surface (or has negligible curvature).

    Returns
    -------
    numpy.ndarray
        The sagitta value(s) corresponding to the input radius/radii.
        Returns axial_z if the radius is close to zero or if it is a fresnel surface.

    Raises
    ------
    ValueError
        If the inputs have invalid shapes or if curvature is negative.

    Notes
    -----
    The sagitta is described by the equation

        .. math::

            s(r) = \frac{cr^2}{1 + \sqrt{1 - c^2r^2\left(1+k\right)}} + \sum_{i=1}A_{i-1}r^{2i} + z_0

    Examples
    --------
    
    .. code-block:: python

        radius = np.array([0.0, 1.0, 2.0])
        curvature = 0.1
        conic_constant = -1.0
        aspheric_coefficients = np.array([0.01, 0.001])
        axial_z = 10.0
        is_fresnel = False
        sag(radius, curvature, conic_constant, aspheric_coefficients, axial_z, is_fresnel)
    
    """

    if not isinstance(aspheric_coefficients, np.ndarray):
        aspheric_coefficients = np.asarray(aspheric_coefficients, dtype=np.float64)

    # Input validation
    if aspheric_coefficients.ndim != 1:
        raise ValueError("aspheric_coefficients must be a 1D array")

    # Ensure radius is an array to allow for consistent indexing, but keep track of original type
    was_list = isinstance(radius, list) 
    was_scalar = isinstance(radius, (int, float))
    original_length = len(radius) if was_list else None
    radius = np.atleast_1d(np.asarray(radius, dtype=np.float64))

    # Handle Fresnel or flat surfaces
    if is_fresnel or np.isclose(curvature, 0, atol=ZERO_TOLERANCE):
        if was_scalar:
            return axial_z
        return np.full_like(radius, axial_z, dtype=np.float64)

    # Aspheric term calculation
    powers = np.arange(1, len(aspheric_coefficients) + 1) * 2
    r_powers = radius[..., np.newaxis]**powers
    aspheric_term = np.sum(aspheric_coefficients * r_powers, axis=-1)

    # Calculate the conic term
    arg_sqrt = 1 - (1 + conic_constant) * curvature**2 * radius**2
    conic_term = curvature * radius**2 / (1 + np.sqrt(np.where(arg_sqrt<0, np.inf, arg_sqrt)))

    result = np.where(np.isclose(radius, 0, atol=ZERO_TOLERANCE), 0, conic_term+aspheric_term) + axial_z

    # Return scalar only if input was a scalar
    if was_scalar:
        result = result[0]
    return result


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass, field
from enum import Enum
import numpy as np
import cupy as cp

from ..visualization._iactsim_style import iactsim_style, MplInteractiveContext

class SurfaceType(Enum):
    """
    Enumeration representing the different types of optical surfaces.

    The directionality of reflection and sensitivity (e.g., "from above" or 
    "from below") is defined within the *local reference frame* of the surface at the point 
    of photon incidence. As a reference:

        - the front side of a surface with negative curvature is the convex side.
        - the front side of a surface with posistive curvature is the concave side.

    Attributes
    ----------
    REFLECTIVE : int
        Represents a reflective surface (both sides).
    REFLECTIVE_FRONT : int
        Represents a surface that is reflective on the front side. 
        In the surface reference frame, only photons arriving with a negative 
        z-component of their direction vector are reflected. 
        The curvature of the surface does not affect this behavior.
    REFLECTIVE_BACK : int
        Represents a surface that is reflective on the back side. In the surface 
        reference frame, only photons arriving with a positive z-component of 
        their direction vector are reflected. The curvature of the surface 
        does not affect this behavior.
    REFRACTIVE : int
        Represents a refractive surface. The refraction is the same on both sides.
    SENSITIVE : int
        Represents the focal plane surface where photons are detected. The sensitivity
        is the same on both sides.
    SENSITIVE_BACK : int
        Represents a surface that is sensitive on the front side. In the surface reference 
        frame, only photons arriving with a negative z-component of their direction 
        vector are detected. The curvature of the surface does not affect 
        this behavior.
    SENSITIVE_FRONT : int
        Represents a surface that is sensitive on the back side. In the surface
        reference frame, only photons arriving with a positive z-component of 
        their direction vector are detected. The curvature of the surface 
        does not affect this behavior.
    OPAQUE : int
        Represents a surface that blocks light.
    DUMMY : int
        Represents a surface that neither reflects nor refracts light. 
        It can be used to introduce artificial absorption or scattering effects, 
        serving as a means to model specific behaviors within the optical system.
    
    """
    REFLECTIVE = 0
    REFLECTIVE_FRONT = 1
    REFLECTIVE_BACK = 2
    REFRACTIVE = 3
    SENSITIVE = 4
    SENSITIVE_BACK = 5
    SENSITIVE_FRONT = 6
    OPAQUE = 7
    DUMMY = 8
    TEST_SENSITIVE = 9
    REFLECTIVE_SENSITIVE = 10

class SensorSurfaceType(Enum):
    """Enumeration representing different types of sensor surfaces.

    Attributes
    ----------
    NONE : int
        Represents no specific sensor surface type.
    SIPM_TILE : int
        Represents a SiPM tile sensor surface.
    
    """
    NONE = 0
    SIPM_TILE = 1

class SurfaceShape(Enum):
    """
    Enumeration representing different types of surface shapes.

    This enum defines common surface shapes encountered in optical systems or
    other fields requiring precise geometrical definitions.

    """
    ASPHERICAL = 0
    """Represents an aspherical surface."""
    CYLINDRICAL = 1
    """Represents a cylindrical surface."""
    FLAT = 2
    """Represents a flat surface."""
    SPHERICAL = 3
    """Represents a spherical surface."""

class ApertureShape(Enum):
    """
    Enumeration representing the possible shapes of an aperture.

    Attributes
    ----------
    CIRCULAR : int
        Represents a circular aperture.
    HEXAGONAL : int
        Represents a flat-top hexagonal aperture.
    SQUARE : int
        Represents a square aperture.
    HEXAGONAL_PT : int
        Represents a pointy-top hexagonal aperture.
    """
    CIRCULAR = 0
    HEXAGONAL = 1
    SQUARE = 2
    HEXAGONAL_PT = 3

@dataclass
class SurfaceTextures:
    """
    Container for GPU texture objects (optical + efficiency) 
    and the metadata required to map physical values (wavelength, angle) 
    to texture coordinates. It can retains a reference to the underlying 
    CUDA arrays to prevent premature garbage collection.
    Because efficiency can be defined on a different grid than transmittance/reflectance,
    this class stores two separate sets of coordinate mappings.
    """
    # Front-side optical textures
    transmittance: cp.cuda.texture.TextureObject
    reflectance: cp.cuda.texture.TextureObject
    
    # Back-side optical textures
    transmittance_back: cp.cuda.texture.TextureObject
    reflectance_back: cp.cuda.texture.TextureObject
    
    # Optical grid
    optical_wavelength_start: float
    optical_wavelength_inv_step: float
    optical_angle_start: float
    optical_angle_inv_step: float

    # Detection efficiency texture
    efficiency: cp.cuda.texture.TextureObject
    
    # Efficiency grid
    efficiency_wavelength_start: float
    efficiency_wavelength_inv_step: float
    efficiency_angle_start: float
    efficiency_angle_inv_step: float

    # References to prevent garbage collection
    _t_array: cp.cuda.texture.CUDAarray = field(default=None, repr=False)
    _r_array: cp.cuda.texture.CUDAarray = field(default=None, repr=False)
    _t_back_array: cp.cuda.texture.CUDAarray = field(default=None, repr=False)
    _r_back_array: cp.cuda.texture.CUDAarray = field(default=None, repr=False)
    _eff_array: cp.cuda.texture.CUDAarray = field(default=None, repr=False)


@dataclass
class SurfaceProperties:
    """
    Represents surface transmittance, reflectance, absorption and detection efficiency
    as a function of photon wavelength and incidence angle.
    """
    # Front-side optical properties
    transmittance: np.ndarray = field(default=None, metadata={'units': 'None'})
    reflectance: np.ndarray = field(default=None, metadata={'units': 'None'})
    absorption: np.ndarray = field(default=None, metadata={'units': 'None'})

    # Back-side optical properties
    transmittance_back: np.ndarray = field(default=None, metadata={'units': 'None'})
    reflectance_back: np.ndarray = field(default=None, metadata={'units': 'None'})
    absorption_back: np.ndarray = field(default=None, metadata={'units': 'None'})

    # Optical grid
    wavelength: np.ndarray = field(default=None, metadata={'units': 'nanometers'})
    incidence_angle: np.ndarray = field(default=None, metadata={'units': 'degrees'})

    # Detection efficiency
    efficiency: np.ndarray = field(default=None, metadata={'units': 'None'})
    efficiency_kind: str = field(default='incident')
    """ 'incident' (relative to incoming photons) or 'absorbed' (relative to absorbed photons). """

    # Efficiency grid
    efficiency_wavelength: np.ndarray = field(default=None, metadata={'units': 'nanometers'})
    efficiency_incidence_angle: np.ndarray = field(default=None, metadata={'units': 'degrees'})

    def __setattr__(self, key, value):
        if value is not None:
            # Convert lists/tuples to arrays automatically
            if key in {'transmittance', 'reflectance', 'absorption', 
                       'transmittance_back', 'reflectance_back', 'absorption_back',
                       'wavelength', 'incidence_angle', 
                       'efficiency', 'efficiency_wavelength', 'efficiency_incidence_angle'}:
                value = np.asarray(value)
        
        super().__setattr__(key, value)
    
    @property
    def is_defined(self):
        try:
            self._validate()
            return True
        except:
            return False

    def _validate_side(self, t_arr, r_arr, a_arr, side_name="Front"):
        """Helper to validate and fill missing optical properties for one side."""
        if sum([x is None for x in [t_arr, r_arr, a_arr]]) > 1:
            raise(ValueError("At least two properties among transmittance, reflectance and absorption must be defined."))

        if t_arr is None: t_arr = 1. - r_arr - a_arr
        if r_arr is None: r_arr = 1. - t_arr - a_arr
        if a_arr is None: a_arr = 1. - t_arr - r_arr
        
        # Check for negative values
        for name, arr in [('Transmittance', t_arr), ('Reflectance', r_arr), ('Absorption', a_arr)]:
            if np.any(np.asarray(arr) < -1e-9):
                raise ValueError(f"{side_name} {name} cannot contain negative values.")
            if np.any(np.asarray(arr) > 1. + 1e-9):
                raise ValueError(f"{side_name} {name} cannot contain values grater than 1.")
            arr[:] = np.clip(arr, 0.0, 1.0)

        # Check energy conservation
        total = t_arr + r_arr + a_arr
        if np.any(total > 1.0 + 1e-9):
            max_val = np.max(total)
            raise ValueError(f"{side_name} energy conservation violation: r + t + a > 1 (max: {max_val:.4f})")
            
        return t_arr, r_arr, a_arr

    def _validate(self):
        """
        Validation of properties. 
        """

        # Validate efficiency
        has_efficiency_grid = (self.efficiency_wavelength is not None or self.efficiency_incidence_angle is not None)
        
        if self.efficiency is not None:
            if not has_efficiency_grid:
                 raise ValueError("Efficiency defined but wavelength/incidence angle are missing.")
            
            # Range check
            if np.any(self.efficiency < -1e-9) or np.any(self.efficiency > 1.0 + 1e-9):
                raise ValueError("Efficiency must be between 0 and 1.")
            self.efficiency = np.clip(self.efficiency, 0.0, 1.0)

        # Validate optical properties
        has_optical_grid = (self.wavelength is not None or self.incidence_angle is not None)

        front_optical_props_present = any(x is not None for x in [self.transmittance, self.reflectance, self.absorption])

        back_optical_props_present = any(x is not None for x in [self.transmittance_back, self.reflectance_back, self.absorption_back])

        if back_optical_props_present or front_optical_props_present:
            if not has_optical_grid:
                 raise ValueError("Optical properties defined but wavelength/incidence angle are missing.")
        
        if front_optical_props_present:
            self.transmittance, self.reflectance, self.absorption = self._validate_side(
                self.transmittance, self.reflectance, self.absorption, side_name="Front"
            )

        if front_optical_props_present and back_optical_props_present:
            self.transmittance_back, self.reflectance_back, self.absorption_back = self._validate_side(
                self.transmittance_back, self.reflectance_back, self.absorption_back, side_name="Back"
            )

    def _prepare_texture_data_for_grid(self, data, wl_arr, ang_arr):
        """Pads and formats data for a specific wavelength/angle grid."""
        n_wl = len(wl_arr) if wl_arr is not None else 1
        n_th = len(ang_arr) if ang_arr is not None else 1

        if data is None:
            arr = np.zeros((n_th, n_wl), dtype=np.float32)
        else:
            arr = np.asarray(data, dtype=np.float32)
            # Check compatibility
            if arr.shape != (n_th, n_wl):
                # Allow reshaping only if the total number of elements matches
                if arr.size == n_th * n_wl:
                    arr = arr.reshape((n_th, n_wl))
                else:
                    raise ValueError(f"Data shape {arr.shape} does not match grid dimensions ({n_th}, {n_wl}).")

        # Wavelength padding
        wl_start = 0.0
        wl_inv_step = 0.0
        if wl_arr is not None and n_wl > 1:
            step = float(wl_arr[1] - wl_arr[0])
            if step != 0:
                wl_inv_step = 1.0 / step
                wl_start = float(wl_arr[0]) - step
                arr = np.pad(arr, ((0, 0), (1, 1)), mode='constant', constant_values=0)
        elif wl_arr is not None and n_wl == 1:
            wl_start = float(wl_arr[0])

        # Angle padding
        th_start = 0.0
        th_inv_step = 0.0
        if ang_arr is not None and n_th > 1:
            step = float(ang_arr[1] - ang_arr[0])
            if step != 0:
                th_inv_step = 1.0 / step
                
                pad_top = 1 
                pad_bottom = 1 
                if float(ang_arr[0]) - step < -1e-5: pad_top = 0
                if float(ang_arr[-1]) + step > 90.0 + 1e-5: pad_bottom = 0

                if pad_top == 1: th_start = float(ang_arr[0]) - step
                else: th_start = float(ang_arr[0])
                
                if pad_top > 0 or pad_bottom > 0:
                    arr = np.pad(arr, ((pad_top, pad_bottom), (0, 0)), mode='constant', constant_values=0)
                    
        elif ang_arr is not None and n_th == 1:
            th_start = float(ang_arr[0])

        return arr, wl_start, wl_inv_step, th_start, th_inv_step

    def get_textures(self) -> SurfaceTextures:
        """
        Creates a SurfaceTextures object.
        """
        self._validate()

        no_front_props = self.transmittance is None and self.reflectance is None
        no_back_props = self.transmittance_back is None and self.reflectance_back is None

        # Optical properties
        # If front(back) properties are missing use the back(front) ones

        if no_back_props and no_front_props:
            # Default dummy grid if no optical props
            # The ray-tracing kernel will check if the grid spacing is > 0 
            wl_grid = self.wavelength if self.wavelength is not None else np.array([0.0])
            ang_grid = self.incidence_angle if self.incidence_angle is not None else np.array([0.0])
            t_data, _, _, _, _ = self._prepare_texture_data_for_grid(None, wl_grid, ang_grid)
            r_data, _, _, _, _ = self._prepare_texture_data_for_grid(None, wl_grid, ang_grid)
            opt_wl_s, opt_wl_inv = 0.0, 0.0 
            opt_th_s, opt_th_inv = 0.0, 0.0
            no_back_props = True
            no_front_props = False

        elif no_back_props or no_front_props:
            input_tran = self.transmittance if not no_front_props else self.transmittance_back
            input_refl = self.reflectance if not no_front_props else self.reflectance_back
            t_data, opt_wl_s, opt_wl_inv, opt_th_s, opt_th_inv = self._prepare_texture_data_for_grid(
                input_tran, self.wavelength, self.incidence_angle
            )
            r_data, _, _, _, _ = self._prepare_texture_data_for_grid(
                input_refl, self.wavelength, self.incidence_angle
            )
        else:
            t_data, opt_wl_s, opt_wl_inv, opt_th_s, opt_th_inv = self._prepare_texture_data_for_grid(
                self.transmittance, self.wavelength, self.incidence_angle
            )
            r_data, _, _, _, _ = self._prepare_texture_data_for_grid(
                self.reflectance, self.wavelength, self.incidence_angle
            )
            t_back_data, _, _, _, _ = self._prepare_texture_data_for_grid(
                self.transmittance_back, self.wavelength, self.incidence_angle
            )
            r_back_data, _, _, _, _ = self._prepare_texture_data_for_grid(
                self.reflectance_back, self.wavelength, self.incidence_angle
            )

        if not no_front_props:
            tex_t, arr_t = self._create_texture_2d(t_data)
            tex_r, arr_r = self._create_texture_2d(r_data)
        
        if not no_back_props:
            tex_t_back, arr_t_back = self._create_texture_2d(t_back_data)
            tex_r_back, arr_r_back = self._create_texture_2d(r_back_data)
        
        if no_back_props and not no_front_props:
            tex_t_back, tex_r_back = tex_t, tex_r
            arr_t_back, arr_r_back = None, None
        
        if no_front_props and not no_back_props:
            tex_t, tex_r = tex_t_back, tex_r_back
            arr_t, arr_r = None, None

        # Efficiency
        if self.efficiency is None:
            # Void texture data (is going to be ignored by the ray-traacing kernel)
            eff_data, _, _, _, _ = self._prepare_texture_data_for_grid(None, [0.], [0.])
            eff_wl_s, eff_wl_inv = 0.0, 0.0
            eff_th_s, eff_th_inv = 0.0, 0.0
        else:
            eff_data, eff_wl_s, eff_wl_inv, eff_th_s, eff_th_inv = self._prepare_texture_data_for_grid(
                self.efficiency, self.efficiency_wavelength, self.efficiency_incidence_angle
            )

        tex_eff, arr_eff = self._create_texture_2d(eff_data)

        return SurfaceTextures(
            transmittance=tex_t, reflectance=tex_r,
            transmittance_back=tex_t_back, reflectance_back=tex_r_back,
            optical_wavelength_start=opt_wl_s, optical_wavelength_inv_step=opt_wl_inv,
            optical_angle_start=opt_th_s, optical_angle_inv_step=opt_th_inv,
            efficiency=tex_eff,
            efficiency_wavelength_start=eff_wl_s, efficiency_wavelength_inv_step=eff_wl_inv,
            efficiency_angle_start=eff_th_s, efficiency_angle_inv_step=eff_th_inv,
            _t_array=arr_t, _r_array=arr_r,
            _t_back_array=arr_t_back, _r_back_array=arr_r_back, _eff_array=arr_eff
        )

    def _create_texture_2d(self, texture_data_np):
        """Internal helper to create a TextureObject from a numpy array."""
        texture_data = texture_data_np.astype(np.float32)
        height, width = texture_data.shape
        
        desc = cp.cuda.texture.ChannelFormatDescriptor(
            32, 0, 0, 0,
            cp.cuda.runtime.cudaChannelFormatKindFloat
        )
        
        cu_array = cp.cuda.texture.CUDAarray(
            desc,
            width,
            height
        )
        cu_array.copy_from(texture_data)
        
        res_desc = cp.cuda.texture.ResourceDescriptor(
            cp.cuda.runtime.cudaResourceTypeArray,
            cuArr=cu_array
        )
        
        tex_desc = cp.cuda.texture.TextureDescriptor(
            addressModes=(cp.cuda.runtime.cudaAddressModeClamp, cp.cuda.runtime.cudaAddressModeClamp),
            filterMode=cp.cuda.runtime.cudaFilterModeLinear,
            readMode=cp.cuda.runtime.cudaReadModeElementType,
            normalizedCoords=False
        )

        return cp.cuda.texture.TextureObject(res_desc, tex_desc), cu_array

    @iactsim_style
    def plot(self, kind: str = 'transmittance', heatmap: bool | None = None, cmap: str = "Spectral"):
        """Plot properties including transmittance, reflectance, absorption, and efficiency."""
        self._validate()

        import matplotlib.pyplot as plt

        kind_lower = kind.lower()
        is_back = 'back' in kind_lower

        if kind_lower.startswith('e'):
            data = self.efficiency
            if data is None: 
                raise ValueError("No efficiency defined.")

            # Determine grid
            pl_wl = self.efficiency_wavelength
            pl_ang = self.efficiency_incidence_angle
            
            label_suffix = " (incident)" if self.efficiency_kind == 'incident' else " (absorbed)"
            title_label = "Efficiency" + label_suffix

        else:
            if kind_lower.startswith('t'): 
                base_attr = 'transmittance'
            elif kind_lower.startswith('r'): 
                base_attr = 'reflectance'
            elif kind_lower.startswith('a'): 
                base_attr = 'absorption'
            else: 
                raise ValueError(f"Unknown plot kind: '{kind}'")
            
            target_attr = f"{base_attr}_back" if is_back else base_attr
            data = getattr(self, target_attr)
            if data is None: 
                raise ValueError(f"{target_attr} is not initialized.")

            pl_wl = self.wavelength
            pl_ang = self.incidence_angle
            title_label = f"{base_attr.title()}" + (" (back)" if is_back else " (front)")

        fig, ax = plt.subplots()
        n_wl = len(pl_wl) if pl_wl is not None else 1
        n_th = len(pl_ang) if pl_ang is not None else 1
        
        if data.ndim == 1:
            if n_th == 1: curves = data.reshape(1, n_wl)
            elif n_wl == 1: curves = data.reshape(n_th, 1)
            else: curves = data.reshape(n_th, n_wl)
        else:
            curves = data.reshape(n_th, n_wl)

        if heatmap is None: heatmap = (n_th > 11 and n_wl > 11)
        if not (pl_wl is not None and pl_ang is not None): heatmap = False

        if heatmap:
            contour = ax.contourf(pl_wl, pl_ang, curves, levels=np.linspace(0, 1, 100), cmap=cmap)
            ax.set_xlabel("Wavelength (nm)")
            ax.set_ylabel("Incidence angle ($\\degree$)")
            cbar = fig.colorbar(contour, ax=ax, boundaries=np.linspace(0, 1, 100))
            cbar.set_label(title_label)
        else:
            show_legend = False
            if n_th <= n_wl:
                x_axis = pl_wl if pl_wl is not None else np.arange(n_wl)
                for i in range(n_th):
                    lbl = f'{pl_ang[i]:.1f} $\\degree$' if pl_ang is not None else None
                    ax.plot(x_axis, curves[i], label=lbl)
                ax.set_xlabel('Wavelength (nm)')
                if n_th > 0:
                    show_legend = True
            else:
                x_axis = pl_ang if pl_ang is not None else np.arange(n_th)
                for i in range(n_wl):
                    lbl = f'{pl_wl[i]:.1f} nm' if pl_wl is not None else None
                    ax.plot(x_axis, curves[:, i], label=lbl)
                ax.set_xlabel('Incidence angle ($\\degree$)')
                if n_wl > 0:
                    show_legend = True

            if show_legend:
                ax.legend()
            
            ax.set_ylabel(title_label)
            ax.grid(which='both')

        return fig, ax

    def save_json(self, filename: str):
        """
        Saves the raw surface attributes to a JSON file.
        """
        import json
        
        # Helper to convert data for JSON
        def to_json_friendly(val):
            if val is None:
                return None
            if isinstance(val, np.ndarray):
                return val.tolist() # Convert numpy -> list
            return val

        # Dictionary of all attributes to save
        data = {
            # Optical
            'transmittance': to_json_friendly(self.transmittance),
            'reflectance':   to_json_friendly(self.reflectance),
            'absorption':    to_json_friendly(self.absorption),
            'transmittance_back': to_json_friendly(self.transmittance_back),
            'reflectance_back':   to_json_friendly(self.reflectance_back),
            'absorption_back':    to_json_friendly(self.absorption_back),
            'wavelength':      to_json_friendly(self.wavelength),
            'incidence_angle': to_json_friendly(self.incidence_angle),
            
            # Efficiency
            'efficiency':                 to_json_friendly(self.efficiency),
            'efficiency_kind':            self.efficiency_kind,
            'efficiency_wavelength':      to_json_friendly(self.efficiency_wavelength),
            'efficiency_incidence_angle': to_json_friendly(self.efficiency_incidence_angle),
        }

        # Filter out None values to keep the file clean
        data = {k: v for k, v in data.items() if v is not None}

        with open(filename, 'w') as f:
            # indent=2 ensures pretty printing
            json.dump(data, f, indent=2)

    @classmethod
    def load_json(cls, filename: str):
        """
        Loads surface attributes from a JSON file.
        """
        import json
        
        with open(filename, 'r') as f:
            data = json.load(f)
            
        obj = cls()
        
        # Helper to restore numpy arrays
        def from_json(key):
            val = data.get(key)
            if val is None:
                return None
            # Strings (efficiency_kind) stay strings, lists become arrays
            if isinstance(val, list):
                return np.array(val, dtype=np.float32)
            return val

        # Optical
        obj.transmittance = from_json('transmittance')
        obj.reflectance   = from_json('reflectance')
        obj.absorption    = from_json('absorption')
        
        obj.transmittance_back = from_json('transmittance_back')
        obj.reflectance_back   = from_json('reflectance_back')
        obj.absorption_back    = from_json('absorption_back')
        
        # Grids
        obj.wavelength      = from_json('wavelength')
        obj.incidence_angle = from_json('incidence_angle')
        
        # Efficiency
        obj.efficiency = from_json('efficiency')
        if 'efficiency_kind' in data:
            obj.efficiency_kind = data['efficiency_kind']
            
        obj.efficiency_wavelength      = from_json('efficiency_wavelength')
        obj.efficiency_incidence_angle = from_json('efficiency_incidence_angle')

        return obj

class FresnelSurfacePropertiesGenerator:
    """
    Handles the calculation and population of SurfaceProperties objects 
    based on Fresnel equations for a given Surface instance.
    """

    @staticmethod
    def _calculate_fresnel(n1, n2, theta_deg, polarization='unpolarized'):
        """
        Perform a vectorized Fresnel calculation for a specific polarization.

        Parameters
        ----------
        n1 : np.ndarray
            Refractive index of the incident medium. Shape must be broadcastable 
            to (N_angles, N_wl).
        n2 : np.ndarray
            Refractive index of the transmission medium. Shape must be broadcastable 
            to (N_angles, N_wl).
        theta_deg : np.ndarray
            Incident angles in degrees. Shape must be broadcastable 
            to (N_angles, N_wl).
        polarization : str, optional
            The polarization mode to compute. Options are:
            - 'unpolarized': Average of s- and p-polarization.
            - 's', 'rs': s-polarization only.
            - 'p', 'rp': p-polarization only.
            Default is 'unpolarized'.

        Returns
        -------
        np.ndarray
            Reflectance matrix R with shape (N_angles, N_wl).

        Raises
        ------
        ValueError
            If an invalid polarization string is provided.
        """
        theta_i = np.radians(theta_deg)
        
        # Snell's law
        sin_theta_t = (n1 / n2) * np.sin(theta_i)

        # Total internal reflection
        tir_mask = np.abs(sin_theta_t) > 1.0

        sin_theta_t_safe = np.where(tir_mask, 0.0, sin_theta_t)
        theta_t = np.arcsin(sin_theta_t_safe)

        # Amplitude coefficients
        cos_theta_i = np.cos(theta_i)
        cos_theta_t = np.cos(theta_t)

        # Initialize result
        R_out = None

        # Calculate s-pol if needed
        if polarization in ['s', 'rs', 'unpolarized']:
            rs_num = n1 * cos_theta_i - n2 * cos_theta_t
            rs_den = n1 * cos_theta_i + n2 * cos_theta_t
            rs = np.divide(rs_num, rs_den, out=np.zeros_like(rs_num), where=rs_den!=0)
            Rs = np.abs(rs)**2
            Rs = np.where(tir_mask, 1.0, Rs)
            
            if polarization in ['s', 'rs']:
                R_out = Rs

        # Calculate p-pol if needed
        if polarization in ['p', 'rp', 'unpolarized']:
            rp_num = n2 * cos_theta_i - n1 * cos_theta_t
            rp_den = n2 * cos_theta_i + n1 * cos_theta_t
            rp = np.divide(rp_num, rp_den, out=np.zeros_like(rp_num), where=rp_den!=0)
            Rp = np.abs(rp)**2
            Rp = np.where(tir_mask, 1.0, Rp)

            if polarization in ['p', 'rp']:
                R_out = Rp

        # Unpolarized
        if polarization == 'unpolarized':
            R_out = 0.5 * (Rs + Rp)

        if R_out is None:
            raise ValueError(f"Invalid polarization mode: '{polarization}'. "
                             "Use 's', 'p', or 'unpolarized'.")

        return R_out

    def generate(self, surface, wavelengths=None, angles=None, polarization='unpolarized', inplace=False):
        """
        Generate a populated SurfaceProperties object.

        Parameters
        ----------
        surface : iactsim.optics.Surface
            The surface object containing `material_in` and `material_out` attributes,
            which must have `get_refractive_index(wavelengths)` methods.
        wavelengths : np.ndarray, optional
            1D array of wavelengths in nm. If None, defaults to `np.arange(200, 901, 1)`.
        angles : np.ndarray, optional
            1D array of incidence angles in degrees. If None, defaults to `np.arange(0, 91, 1)`.
        polarization : {'unpolarized', 's', 'p', 'rs', 'rp'}, optional
            The specific polarization to compute for reflectance and transmittance.
            Default is 'unpolarized'.
        inplace : bool, optional
            If True, modifies the provided surface properties in place.
            Default is False.

        Returns
        -------
        iactsim.optics.SurfaceProperties
            If `inplace` is False, a populated properties object containing the calculated matrices:
            - `reflectance`, `transmittance` (Front interface)
            - `reflectance_back`, `transmittance_back` (Back interface)
            - `wavelength`, `incidence_angle` vectors.
        """
        pol_clean = polarization.lower().strip()

        # Apply defaults
        if wavelengths is None:
            wavelengths = np.arange(200, 901, 1)
        
        if angles is None:
            angles = np.arange(0, 91, 1)

        # Ensure inputs are numpy arrays
        wl = np.asarray(wavelengths)
        ang = np.asarray(angles)

        # Retrieve materials and reshape for broadcasting
        # n: (1, N_wl) | ang: (N_ang, 1)
        n_in = surface.material_in.get_refractive_index(wl).reshape(1, -1)
        n_out = surface.material_out.get_refractive_index(wl).reshape(1, -1)
        ang_reshaped = ang.reshape(-1, 1)

        # Front interface (n_in -> n_out)
        R_front = self._calculate_fresnel(n_in, n_out, ang_reshaped, pol_clean)
        
        # Cleanup
        nan_mask_f = np.isnan(R_front)
        T_front = 1.0 - R_front

        # NaNs will be treated as absorption
        R_front[nan_mask_f] = 0.0
        T_front[nan_mask_f] = 0.0

        # Back interface (n_out -> n_in)
        R_back = self._calculate_fresnel(n_out, n_in, ang_reshaped, pol_clean)
        
        # Cleanup
        nan_mask_b = np.isnan(R_back)
        T_back = 1.0 - R_back

        # NaNs will be treated as absorption
        R_back[nan_mask_b] = 0.0
        T_back[nan_mask_b] = 0.0

        # Populate Object
        surface_prop = SurfaceProperties()
        surface_prop.wavelength = wl
        surface_prop.incidence_angle = ang
        
        surface_prop.reflectance = R_front
        surface_prop.transmittance = T_front
        surface_prop.reflectance_back = R_back
        surface_prop.transmittance_back = T_back

        if inplace:
            surface.properties = surface_prop
        else:
            return surface_prop


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

import re
from functools import wraps
import textwrap

def wrap_cupy_rawkernel(raw_kernel, docstring):
    """
    Generates wrapper function for a CuPy RawKernel.
    The wrapper allows you to provide a custom docstring.

    Parameters
    ----------
        raw_kernel: cupy.RawKernel
            The cupy.RawKernel object to wrap.
        docstring: str
            A custom docstring for the wrapped kernel.

    Returns
    -------
        A Python function that wraps the CuPy RawKernel.
    """

    @wraps(raw_kernel)  # Preserve original kernel's metadata
    def wrapper(*args, **kwargs):
        raw_kernel(*args, **kwargs)

    wrapper.__doc__ = docstring

    return wrapper

def clean(string):
    string_clean = re.sub(r"^ \* ?", "", string, flags=re.MULTILINE).strip()
    if string_clean.startswith("*"):
        string_clean = string_clean[1:]
    return string_clean

def convert_doxygen_to_numpy(cuda_source, function_name):
    """
    Converts Doxygen-style documentation within the given source code to NumPy-style docstrings.

    Parameters
    ----------
        cuda_source: str
            The source code.
        function_name: str
            The name of the function.

    Returns
    -------
        docstring: str 
            The NumPy-style docstring or None if not found.
    """

    # Find all functions with docstring
    function_pattern = re.compile(
        r"/\*\*\s*\n\s*\*\s*@brief\s*(.*?)\n(.*?)\*/\s*(__global__|__device__)\s+.*?\s+(\w+)\s*\(",
        re.DOTALL,
    )

    # Cicly over all functions
    for match in function_pattern.finditer(cuda_source):

        # Look for the specified function
        current_function_name = match.group(4)
        if current_function_name != function_name:
            continue
        
        is_global = match.group(3) == '__global__'

        # Isolate the function source code
        start = match.start()

        # Find the *correct* closing parenthesis for the parameter list
        open_paren_index = match.end()  # start after the opening parenthesis
        brace_count = 1
        cur_pos = open_paren_index
        while brace_count > 0 and cur_pos < len(cuda_source):
            if cuda_source[cur_pos] == '(':
                brace_count += 1
            elif cuda_source[cur_pos] == ')':
                brace_count -= 1
            cur_pos += 1

        end = cur_pos  # end after the closing parenthesis
        function_source = cuda_source[start:end]

        # Remove comments
        function_source = re.sub(r'//.*', '', function_source)
        
        param_string_start = [m.start() for m in re.finditer(r'\(',function_source)][-1]+1
        param_string_end = [m.start() for m in re.finditer(r'\)',function_source)][-1]
        param_types_str = function_source[param_string_start:param_string_end].strip()
        param_types_str = re.sub(r"[\n\t ]+", " ", param_types_str)
        
        # Parse parameter types and names
        param_types_dict = {}
        if param_types_str:
          params = [p.strip() for p in param_types_str.split(",") if p.strip()]
          for param in params:
            param = re.sub(r'//.*', '', param).strip()
            if not param:
                continue
            
            parts = param.rsplit(" ", 1)
            if len(parts) == 2:
              param_type, param_name = parts[0].strip(), parts[1].strip()
              param_type = param_type.replace(' *', '*').replace(' &', '&')
              if param_name.startswith('*') or param_name.startswith('&'):
                param_type += param_name[0]
                param_name = param_name[1:]

              param_types_dict[param_name] = param_type
            elif len(parts) == 1:
              param_types_dict[parts[0]] = "Unknown"

        # Isolate function description removing all doxygen directives
        brief = match.group(1).strip()
        detailed = match.group(2).strip()

        # Isolate parameter names
        param_pattern = re.compile(r"@param\s+(\w+)\s+((?:(?!@param|@warning|@note|@return|@returns|@see|@code|@details).)+)", re.DOTALL)
        params = param_pattern.findall(detailed)

        # Handle doxygen directives
        warnings = re.findall(r'@warning\s+((?:(?!@param|@warning|@note|@return|@returns|@see|@code|@details).)+)', detailed, re.DOTALL)
        notes = re.findall(r'@note\s+((?:(?!@param|@warning|@note|@return|@returns|@see|@code|@details).)+)', detailed, re.DOTALL)
        returns = re.findall(r'@return(?:s)?\s+((?:(?!@param|@warning|@note|@return|@returns|@see|@code|@details).)+)', detailed, re.DOTALL)
        see_also = re.findall(r'@see\s+((?:(?!@param|@warning|@note|@return|@returns|@see|@code|@details).)+)', detailed, re.DOTALL)
        code_blocks = re.findall(r'@code(?:cpp)?\s*([\s\S]*?)\s*@endcode', detailed)
        details = re.findall(r'@details\s+((?:(?!@param|@warning|@note|@return|@returns|@see|@code|@details).)+)', detailed, re.DOTALL)

        # Clean description
        detailed_clean = re.sub(r"@param\s+\w+\s+(?:(?!@param|@warning|@note|@return|@returns|@see|@code|@details).)+", "", detailed, flags=re.DOTALL).replace('*', '').strip()
        detailed_clean = re.sub(r'@warning\s+((?:(?!@param|@warning|@note|@return|@returns|@see|@code|@details).)+)',"",detailed_clean, flags=re.DOTALL).strip()
        detailed_clean = re.sub(r'@note\s+((?:(?!@param|@warning|@note|@return|@returns|@see|@code|@details).)+)',"",detailed_clean, flags=re.DOTALL).strip()
        detailed_clean = re.sub(r'@return(?:s)?\s+((?:(?!@param|@warning|@note|@return|@returns|@see|@code|@details).)+)',"",detailed_clean, flags=re.DOTALL).strip()
        detailed_clean = re.sub(r'@see\s+((?:(?!@param|@warning|@note|@return|@returns|@see|@code|@details).)+)',"",detailed_clean, flags=re.DOTALL).strip()
        detailed_clean = re.sub(r'@code(?:cpp)?\s*([\s\S]*?)\s*@endcode',"",detailed_clean, flags=re.DOTALL).strip()
        detailed_clean = re.sub(r'@details\s+((?:(?!@param|@warning|@note|@return|@returns|@see|@code|@details).)+)',"",detailed_clean, flags=re.DOTALL).strip()
        detailed_clean = re.sub(r'\s+', ' ', detailed_clean).strip()

        # Clean brief description
        brief = re.sub(r'\s+', ' ', brief).strip()
        
        # Write docstring
        numpy_docstring = f"{brief.replace('*', '').strip()}\n\n"
        numpy_docstring += f"{detailed_clean}\n\n"

        if details:
            # numpy_docstring += "\n\n---------\n\n"
            details[0] = details[0]
            for detail_text in details:
                detail_text_clean = clean(detail_text)
                numpy_docstring += f"{textwrap.dedent(detail_text_clean.strip())}\n\n"

        if params:
            numpy_docstring += "Parameters\n----------\n"
            for param_name, param_desc in params:
                param_type = param_types_dict.get(param_name, "Unknown")
                param_desc_clean = clean(param_desc)#param_desc.replace('*', '').strip()
                param_desc_clean = param_desc_clean.replace('\n', ' ').strip()
                param_desc_clean = re.sub(r'\s+', ' ', param_desc_clean).strip()
                numpy_docstring += f"{param_name} : {param_type}\n    {param_desc_clean}\n"

        if returns:
            numpy_docstring += "\nReturns\n-------\n"
            for ret_desc in returns:
                # ret_desc_clean = ret_desc.replace('*', '').strip()
                ret_desc_clean = clean(ret_desc)
                numpy_docstring += f"{ret_desc_clean}\n"
        
        if warnings:
            numpy_docstring += "\nWarnings\n--------\n"
            for warning in warnings:
                warning_clean = clean(warning)
                # warning_clean = warning.replace('*', '').strip()
                numpy_docstring += f"{warning_clean}\n"
        
        if notes:
            numpy_docstring += "\nNotes\n-----\n"
            for note_text in notes:
                note_text_clean = clean(note_text)#note_text.replace('*','').strip()
                numpy_docstring += f"{note_text_clean}\n"
        
        if see_also:
            numpy_docstring += "\nSee Also\n--------\n"
            for see in see_also:
                see_clean = clean(see)#see.replace('*', '').strip()
                numpy_docstring += f"{see_clean}\n"
        
        if code_blocks:
            numpy_docstring += "\nExamples\n--------\n"
            for code_text in code_blocks:
                clean_code_text = clean(code_text)#code_text.replace('*', '')
                numpy_docstring += ".. code-block:: cpp\n\n    "
                numpy_docstring += ''.join([line+'\n' for line in clean_code_text.split('\n ')])
                numpy_docstring += "\n"
        
        if is_global:
            numpy_docstring += "\nAttention\n---------\n"
            numpy_docstring += "This is a wrapped CuPy `RawKernel` object,\n"
            numpy_docstring += "you can launch it as follows:\n\n"
            numpy_docstring += ".. code-block:: python\n\n"
            numpy_docstring += "   blocks = ...  # Number of blocks\n"
            numpy_docstring += "   threads = ...  # Number of threads per block\n"
            numpy_docstring += "   shared_memory = ...  # Dynamic shared memory buffer size (if needed)\n"
            numpy_docstring += f"   args = (...)  # Provide the arguments\n"
            numpy_docstring += "   kernel(grid, block, args, shared_mem=shared_memory)\n\n"
            numpy_docstring += "If you are new to CuPy, please see the CuPy `documentation`_.\n\n"
            numpy_docstring += ".. _documentation: https://cupy.dev/\n"

        return numpy_docstring

    return None


def get_kernel(module, kernel_name):
    """Get kernel wrapper.

    Parameters
    ----------
    module : cupy.RawModule
        A CuPy rawModule
    kernel_name : str
        Name of the kernel to retrieve and wrap.
    
    Return
    ------
        Wrapper to the RawKernel with a NumPy-style docstring.
    """
    docstring = convert_doxygen_to_numpy(module.code, kernel_name)
    return wrap_cupy_rawkernel(module.get_function(kernel_name), docstring)


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

# Use module __getattr__ and __dir__ for lazy submodule imports.
# Follwing PEP 562 (https://peps.python.org/pep-0562/)

import importlib as _importlib

try:
    from iactsim import utils
    from iactsim import _version
    __version__ = _version.version
except ImportError as e:
    msg = "The `iactsim` install you are using seems to be broken, " + \
            "(extension modules cannot be imported), " + \
            "please try reinstalling."
    raise ImportError(msg) from e

submodules = [
    'electronics',
    'optics',
    'utils',
    'visualization',
    'io',
    'models',
    'iactxx',
]

main = {
    'IACT': 'iactsim._iact',
}

__all__ = submodules + list(main)

def __dir__():
    return __all__

def __getattr__(name):
    if name in submodules:
        return _importlib.import_module(f'iactsim.{name}')
    elif name in main:
        module = _importlib.import_module(main[name])
        return getattr(module, name)
    else:
        raise AttributeError(f"Module 'iactsim' has no attribute '{name}'")


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

import vtk
import math
import numpy as np

from ..optics._surface_misc import ApertureShape

def hexagonal_clipping(half_aperture, offset, structuredGrid, inside, input_connection=None, pointy_top=False):
    """Clip an hexagonal aspheric surface with option for orientation.

    Parameters
    ----------
    half_aperture : float
        Surface half aperture (defines the apothem/distance to faces).
    offset : tuple(float, float)
        Segment offset (x, y).
    structuredGrid : vtk.StructuredGrid
        The surface to clip.
    inside : bool
        Whether to keep the surface inside or outside the half aperture.
    input_connection : vtkClipDataSet, optional
        Apply the clipping to another clipped surface, by default None.
    pointy_top : bool, optional
        If True, orients the hexagon with a pointy top (normals at 0, 60...).
        If False (default), orients with a flat top (normals at 30, 90...).

    Returns
    -------
    vtkClipDataSet
        Clipped surface.
    """
    hex_apothem = half_aperture
    booleanOp = vtk.vtkImplicitBoolean()
    booleanOp.SetOperationTypeToIntersection()
    num_sides = 6
    angle_step = 2.0 * math.pi / num_sides
    
    # Determine the phase shift based on orientation
    # Flat-top: shift by half step (30 degrees) so normals are at 30, 90, etc.
    # Pointy-top: 0 shift so normals are at 0, 60, etc.
    phase = 0.0 if pointy_top else 0.5 * angle_step

    for i in range(num_sides):
        angle = i * angle_step + phase
        
        nx_plane, ny_plane = math.cos(angle), math.sin(angle)
        ox_plane, oy_plane = hex_apothem * nx_plane, hex_apothem * ny_plane
        
        plane = vtk.vtkPlane()
        # Apply the offset to the origin of the plane
        plane.SetOrigin(ox_plane + offset[0], oy_plane + offset[1], 0.0)
        plane.SetNormal(nx_plane, ny_plane, 0.0)
        booleanOp.AddFunction(plane)
    
    clipper = vtk.vtkClipDataSet()
    if input_connection is None:
        clipper.SetInputData(structuredGrid)
    else:
        clipper.SetInputConnection(input_connection.GetOutputPort())
        
    clipper.SetClipFunction(booleanOp)
    clipper.SetInsideOut(inside)
    clipper.SetValue(0.0)
    clipper.Update()
    
    return clipper

def square_clipping(half_aperture, offset, structuredGrid, inside, input_connection=None):
    """Clip a square aspheric surface.

    Parameters
    ----------
    half_aperture : float
        Surface half aperture.
    offset : float, float
        Segment offset.
    structuredGrid : vtk.StructuredGrid
        The surface to clip.
    inside : bool
        Whether to keep the surface inside or outside the half aperture.
    input_connection : vtkClipDataSet, optional
        Apply the clipping to another clipped surface, by default None.

    Returns
    -------
    vtkClipDataSet
        Clipped surface.
    """
    hex_apothem = half_aperture
    booleanOp = vtk.vtkImplicitBoolean()
    booleanOp.SetOperationTypeToIntersection()
    num_sides = 4
    angle_step = 2.0 * math.pi / num_sides
    for i in range(num_sides):
        angle = i * angle_step + angle_step
        nx_plane, ny_plane = math.cos(angle), math.sin(angle)
        ox_plane, oy_plane = hex_apothem * nx_plane, hex_apothem * ny_plane
        plane = vtk.vtkPlane()
        plane.SetOrigin(ox_plane+offset[0], oy_plane+offset[1], 0.0)
        plane.SetNormal(nx_plane, ny_plane, 0.0)
        booleanOp.AddFunction(plane)
    
    clipper = vtk.vtkClipDataSet()
    if input_connection is None:
        clipper.SetInputData(structuredGrid)
    else:
        clipper.SetInputConnection(input_connection.GetOutputPort())
    clipper.SetClipFunction(booleanOp)
    clipper.SetInsideOut(inside)
    clipper.SetValue(0.0)
    clipper.Update()
    return clipper

def circular_clipping(half_aperture, offset, structuredGrid, inside, input_connection=None):
    """Clip a circular aspheric surface.

    Parameters
    ----------
    half_aperture : float
        Surface half aperture.
    offset : float, float
        Segment offset.
    structuredGrid : vtk.StructuredGrid
        The surface to clip.
    inside : bool
        Whether to keep the surface inside or outside the half aperture.
    input_connection : vtkClipDataSet, optional
        Apply the clipping to another clipped surface, by default None.

    Returns
    -------
    vtkClipDataSet
        Clipped surface.
    """

    # Aperture clipping
    cylinder = vtk.vtkCylinder()
    cylinder.SetCenter(offset[0], offset[1], 0.0)
    cylinder.SetAxis(0.0, 0.0, 1.0)
    cylinder.SetRadius(half_aperture)

    # Create the clipper filter
    clipper = vtk.vtkClipDataSet()
    if input_connection is None:
        clipper.SetInputData(structuredGrid)
    else:
        clipper.SetInputConnection(input_connection.GetOutputPort())
    clipper.SetClipFunction(cylinder)
    clipper.SetInsideOut(inside)
    clipper.SetValue(0.0)
    clipper.GenerateClippedOutputOn()
    clipper.Update()
    return clipper

def create_aspheric_surface_actor(surface, res=None):
    """Create an actor for an aspheric surface.

    Parameters
    ----------
    surface : iactsim.optics.AsphericSurface
        The optical surface.
    res : float, optional
        Mesh resolution in mm, by default 10 mm.

    Returns
    -------
    vtkActor
        The surface actor.
    """
    r_max = surface.half_aperture

    if res is None:
        res = r_max / 50
    
    if surface.aperture_shape in [ApertureShape.HEXAGONAL, ApertureShape.HEXAGONAL_PT]:
        r_max *= 1.1547
    
    if surface.aperture_shape == ApertureShape.SQUARE:
        r_max *= 1.4142
    
    # Meshgrid
    n = int(r_max/res)
    x_coords = np.linspace(surface.offset[0]-r_max, surface.offset[0]+r_max, n)
    y_coords = np.linspace(surface.offset[1]-r_max, surface.offset[1]+r_max, n)
    X, Y = np.meshgrid(x_coords, y_coords, indexing='xy')

    # Radius
    R = np.sqrt(X**2 + Y**2)

    # TODO: Fix this including the offset
    # Check for value outside of the aperture
    # mask_valid = R <= surface.half_aperture
    R_calc = R#np.where(mask_valid, R, r_max)
    X_calc = X#np.where(mask_valid, X, X*r_max/R)
    Y_calc = Y#np.where(mask_valid, Y, Y*r_max/R)

    Z = surface.sagitta(R_calc)

    # Populate vtkStructuredGrid
    points = vtk.vtkPoints()
    for j in range(n):
        for i in range(n):
            points.InsertNextPoint(X_calc[j, i], Y_calc[j, i], Z[j, i])
    structuredGrid = vtk.vtkStructuredGrid()
    structuredGrid.SetDimensions(n, n, 1)
    structuredGrid.SetPoints(points)

    aperture_is_hexagonal = surface.aperture_shape in [ApertureShape.HEXAGONAL, ApertureShape.HEXAGONAL_PT]
    hole_is_hexagonal = surface.central_hole_shape in [ApertureShape.HEXAGONAL, ApertureShape.HEXAGONAL_PT]

    if aperture_is_hexagonal:
        pointy_top = surface.aperture_shape == ApertureShape.HEXAGONAL_PT
        clipper = hexagonal_clipping(surface.half_aperture, surface.offset, structuredGrid, True, pointy_top=pointy_top)
    elif surface.aperture_shape == ApertureShape.CIRCULAR:
        clipper = circular_clipping(surface.half_aperture, surface.offset, structuredGrid, True)
    elif surface.aperture_shape == ApertureShape.SQUARE:
        clipper = square_clipping(surface.half_aperture, surface.offset, structuredGrid, True)
    else:
        raise(RuntimeError(f"Aperture shape '{surface.aperture_shape}' not supported."))

    if hole_is_hexagonal:
        pointy_top = surface.aperture_shape == ApertureShape.HEXAGONAL_PT
        clipper = hexagonal_clipping(surface.central_hole_half_aperture, surface.offset, structuredGrid, False, clipper, pointy_top)
    elif surface.central_hole_shape == ApertureShape.CIRCULAR:
        clipper = circular_clipping(surface.central_hole_half_aperture, surface.offset, structuredGrid, False, clipper)
    elif surface.central_hole_shape == ApertureShape.SQUARE:
        clipper = square_clipping(surface.central_hole_half_aperture, surface.offset, structuredGrid, False, clipper)
    else:
        raise(RuntimeError(f"Aperture shape '{surface.aperture_shape}' not supported."))
    
    transform_off = vtk.vtkTransform()
    transform_off.Translate(-surface.offset[0],-surface.offset[1],-surface.sagitta(math.sqrt(surface.offset[0]**2+surface.offset[1]**2)))
    transformFilter = vtk.vtkTransformFilter()
    transformFilter.SetInputConnection(clipper.GetOutputPort())
    transformFilter.SetTransform(transform_off)
    transformFilter.Update()
    
    # Add Scalar Data
    # scalars = vtk.vtkDoubleArray()
    # scalars.SetName("Radius")
    # scalars.SetNumberOfTuples(nx * ny)
    # k = 0
    # for j in range(ny):
    #     for i in range(nx):
    #         scalars.SetValue(k, R[j,i])
    #         k += 1
    # structuredGrid.GetPointData().SetScalars(scalars)

    # # The output of the clipper is a vtkUnstructuredGrid
    # clipped_output = clipper.GetOutput()

    mapper = vtk.vtkDataSetMapper()
    if clipper is None:
        mapper.SetInputData(structuredGrid)
    else:
        # mapper.SetInputConnection(clipper.GetOutputPort())
        mapper.SetInputConnection(transformFilter.GetOutputPort())
    
    mapper.ScalarVisibilityOff()
    
    actor = vtk.vtkActor()
    actor.SetMapper(mapper)

    return actor

def create_cylindric_surface_actor(surface, res=None):
    """create a VTK actor for a cylindrical surface.

    Parameters
    ----------
    surface : CylindricalSurface
        A cylindrical surface.
    res : float, optional
        Resolution (in mm) used to compute the number of facets, 60 facets are used.

    Returns
    -------
    vtk.vtkActor
        Actor representing the cylinder.
    """
    capping = surface.top and surface.bottom
    cylinder = vtk.vtkCylinderSource()
    cylinder.SetCenter(0, 0, 0)
    cylinder.SetRadius(surface.radius)
    cylinder.SetHeight(surface.height)
    if res is not None:
        res = int(2*np.pi / (2*np.arcsin(res/2/surface.radius)))
    else:
        res = 60
    cylinder.SetResolution(res)
    cylinder.SetCapping(capping)
    cylinder.Update()

    appendFilter = None
    if surface.top != surface.bottom:
        capDisk = vtk.vtkDiskSource()
        capDisk.SetInnerRadius(0.0)
        capDisk.SetOuterRadius(cylinder.radius)
        capDisk.SetCircumferentialResolution(cylinder.resolution)
        capDisk.Update()
        transform_cap = vtk.vtkTransform()
        if surface.top:
            z = 0.5*cylinder.height
        else:
            z = -0.5*cylinder.height
        transform_cap.Translate(0, z, 0)
        transform_cap.RotateX(-90)
        transform_cap.Update()
        
        transform_filter_cap = vtk.vtkTransformPolyDataFilter()
        transform_filter_cap.SetInputConnection(capDisk.GetOutputPort())
        transform_filter_cap.SetTransform(transform_cap)
        transform_filter_cap.Update()

        appendFilter = vtk.vtkAppendPolyData()
        appendFilter.AddInputConnection(cylinder.GetOutputPort())
        appendFilter.AddInputConnection(transform_filter_cap.GetOutputPort())
        appendFilter.Update()
    
    transform_cyl = vtk.vtkTransform()
    transform_cyl.RotateX(90)
    transformFilter = vtk.vtkTransformPolyDataFilter()
    if appendFilter is None:
        transformFilter.SetInputConnection(cylinder.GetOutputPort())
    else:
        transformFilter.SetInputConnection(appendFilter.GetOutputPort())
    transformFilter.SetTransform(transform_cyl)
    transformFilter.Update()
    
    cylinderMapper = vtk.vtkPolyDataMapper()
    cylinderMapper.SetInputData(transformFilter.GetOutput())
    
    actor = vtk.vtkActor()
    actor.SetMapper(cylinderMapper)
    
    return actor

def quit(render_window_interactor):
    render_window = render_window_interactor.GetRenderWindow()
    render_window.Finalize()
    render_window_interactor.TerminateApp()


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

import os as os
from pathlib import Path as Path
import cupy as cp
import numpy as np
import struct
from typing import Tuple, Optional, Union

from ...utils._kernels import get_kernel

path = Path(__file__).parent / "poisson_flash.cu"
with open(path) as _source_file:
    source_code = _source_file.read()

include_path = ''.join([os.environ['CPATH']])
module = cp.RawModule(code=source_code, backend='nvcc', options=(''.join(['-I',include_path]),'--use_fast_math', '-std=c++11', '--extra-device-vectorization', '-O3'))
generate_map = module.get_function("generate_map")
generate_arrival_times = module.get_function("generate_arrival_times")

def generate_poisson_flash(
        n_pixels: int,
        mean: Union[cp.ndarray, np.ndarray],
        t_start: float,
        duration: float,
        seed: Optional[int] = None) -> Tuple[cp.ndarray, cp.ndarray]:
    """Generate Poisson arrival times in a time window equal for all pixels.

    This function simulates the arrival times of photo-electrons
    generated by a Poisson process, representing a light flash.  It assumes
    the same time window (start time and duration) for all pixels, but allows
    for a different mean number of photo-electrons for each pixel.

    Parameters
    ----------
    n_pixels : int
        Total number of pixels.
    mean : Union[cp.ndarray, numpy.ndarray]
        Mean number of photo-electrons for each pixel. This should be a
        1D array of length `n_pixels`.
    t_start : float
        Start time of the flash (in ns).
    duration : float
        Duration of the flash (in ns).
    seed : int, optional
        Random seed for reproducibility. If None, a random seed is generated.

    Returns
    -------
    Tuple[cp.ndarray, cp.ndarray]
        A tuple containing:

        - **pes**: A 1D CuPy array of shape (total_n_pes,) containing the
          arrival times of all photo-electrons.
        
        - **pes_map**: A 1D CuPy array of shape (n_pixels + 1,) representing
          a mapping to access the arrival times for each pixel.  The arrival
          times for pixel `i` are given by ```pes[pes_map[i]:pes_map[i+1]]```.

    Examples
    --------
    .. code-block:: python

        import cupy as cp
        n_pixels = 10
        mean = cp.array([2.0, 3.5, 1.0, 4.2, 0.8, 2.7, 1.9, 3.1, 2.5, 1.3])
        t_start = 0.0
        duration = 10.0
        pes, pes_map = generate_poisson_flash(n_pixels, mean, t_start, duration)
        # Access arrival times for pixel 3:
        pixel_3_times = pes[pes_map[3]:pes_map[4]]

    """
    if seed is None:
        seed = struct.unpack('>I', os.urandom(4))[0]

    mean = cp.asarray(mean, dtype=cp.float32)
    
    pe_per_pixel = cp.empty((n_pixels+1,), dtype=cp.int32)
    
    blocksize = 128
    nblock = int(np.ceil(n_pixels/blocksize))
    generate_map((nblock,), (blocksize,), (pe_per_pixel,mean, cp.uint64(seed), cp.int32(n_pixels)))

    pe_map = cp.cumsum(pe_per_pixel, dtype=cp.int32)

    n_pes = pe_map[-1].get()
    pe_arrival_times = cp.empty((n_pes,), dtype=cp.float32)

    t_start = cp.float32(t_start)
    duration = cp.float32(duration)
    
    blocksize = 128
    nblock = int(np.ceil(n_pes/blocksize))
    generate_arrival_times((nblock,), (blocksize,), (pe_arrival_times, t_start, duration, cp.uint64(seed), cp.int32(n_pes)))

    return pe_arrival_times, pe_map


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

_functions = {
    'scatter': 'iactsim.visualization._scatter_plot',
    'histogram2d': 'iactsim.visualization._histogram2d',
    'plot_sipm_module_pixels': 'iactsim.visualization._sipm_camera_plots',
    'plot_sipm_modules': 'iactsim.visualization._sipm_camera_plots',
    'VTKOpticalSystem': 'iactsim.visualization._vtk_optical_system',
}

_all = _functions

import importlib as _importlib

def __getattr__(name):
    if name not in _all:
        raise AttributeError(f"Module 'iactsim.visualization' has no attribute '{name}'")
    
    _sub_module = _importlib.import_module(_all[name])

    return getattr(_sub_module, name)

__all__ = list(_all) + [s for s in dir() if not s.startswith('_')]
def __dir__():
    return __all__


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import vtk

from ._vtk_utils import (
    create_aspheric_surface_actor,
    create_cylindric_surface_actor,
    quit
)

from ..optics._surface_misc import SurfaceType, SurfaceShape

from ..optics._cpu_transforms import local_to_telescope_rotation, local_to_pointing_rotation

from .._iact import IACT

from ..optics._optical_system import OpticalSystem


class VTKOpticalSystem():
    """Class to viasualize the geometry of an optical system.
    Each surface actor can be accessed from :py:attr:`actors` attribute after a :py:meth:`update` call.

    Parameters
    ----------
    optical_system : OpticalSystem or IACT
        Optical sistem for which visualize surfaces.
    resolution : float, optional
            Objects mesh resolution (in mm). By default 10 mm.

    Notes
    -----
    If you perform some operation on an actor, make sure to call :py:meth:`start_render` with ``update=False``, otherwise the actors will be replaced.
    
    """
    def __init__(self, optical_system, resolution=None):
        self.actors = {}
        """Dictionary of surface name: surface actor."""

        if issubclass(type(optical_system), OpticalSystem):
            self.os = optical_system
            self._tel_to_local_rot =  np.eye(3)
            self._translation_vector = np.zeros(3)
        elif issubclass(type(optical_system), IACT):
            self.os = optical_system.optical_system
            self._tel_to_local_rot =  local_to_telescope_rotation(*optical_system.pointing).T
            self._translation_vector = optical_system.position
        else:
            raise(ValueError("optical system must be an instance of OpticalSystem or IACT."))

        """Optical system to be visualized."""
        
        self._default_surface_color = (0.7, 0.7, 0.75)

        self.surface_type_colors = {
            SurfaceType.REFRACTIVE: (0.6, 0.4, 1),
            SurfaceType.OPAQUE: (1, 0.3, 0.3)
        }

        self.surface_type_opacity = {
            SurfaceType.REFRACTIVE: 0.075,
        }
        
        self.surface_type_specular = {
            SurfaceType.REFRACTIVE: 0.2,
            SurfaceType.REFLECTIVE: 0.8,
            SurfaceType.REFLECTIVE_SENSITIVE: 0.6
        }
        

        """Dictionary to custumize surface colors based on surface type."""

        # Wireframe
        self.wireframe = False
        """Whether to use wireframe representation by default."""

        # Window size
        self.window_size = (1024,1024)
        """Window size in pixel."""

        self._resolution = resolution # mm
        """Objects mesh resolution (in mm)."""

        self._ray_opacity = None
        """Ray opacity."""

        self._update()

        self._apply_global_transform()
    
    def _update(self):
        """Generate all surface actors.
        """
        self.actors = {}
        for i,s in enumerate(self.os):
            if s._shape == SurfaceShape.CYLINDRICAL:
                actor = create_cylindric_surface_actor(s, self._resolution)
            else:
                actor = create_aspheric_surface_actor(s, self._resolution)
                

            transform = vtk.vtkTransform()
            transform.Translate(*s.position)
            R = s.get_rotation_matrix()
            vtk_rotation_matrix = vtk.vtkMatrix4x4()
            for i in range(3):
                for j in range(3):
                    vtk_rotation_matrix.SetElement(i, j, R[j, i])
            transform.PreMultiply()
            transform.Concatenate(vtk_rotation_matrix)
            transform.Update()
            actor.SetUserTransform(transform)
            
            if self.wireframe:
                actor.GetProperty().SetRepresentationToWireframe()
            
            if s.type in self.surface_type_colors:
                color = self.surface_type_colors[s.type]
            else:
                color = self._default_surface_color
            
            if s.type in self.surface_type_opacity:
                opacity = self.surface_type_opacity[s.type]
                actor.GetProperty().SetOpacity(opacity)
            else:
                actor.GetProperty().SetOpacity(1.0)

            actor.GetProperty().SetColor(*color)
            actor.GetProperty().SetAmbient(0.3)
            actor.GetProperty().SetDiffuse(0.5)
            actor.GetProperty().SetSpecular(0.2)
            actor.GetProperty().SetSpecularPower(10)

            self.actors[s.name] = actor

    @staticmethod
    def _get_sigmoid_opacity(n_rays, n_50=200, k=1.0, min_opacity=0.01):
        """
        Calculates ray opacity using a descending sigmoid curve.
        
        Parameters
        ----------
        n_rays : int
            Number of rays being plotted.
        n_50 : int
            The 'knee' of the curve: number of rays where opacity is ~0.5.
        k : float
            Steepness of the curve (the greater the steeper the curve). 
        min_opacity : float
            Minimum visibility floor.
        """
        if n_rays <= 0: return 0.0
        
        # Calculate sigmoid
        opacity = min_opacity + (1.0 - min_opacity) / (1 + (n_rays / n_50)**k)
        
        return opacity

    def start_render(self, camera_position=None, focal_point=None, view_up=(0, 0, 1), orthographic=False):
        """Render the optical system geometry on a VTK window.

        Parameters
        ----------
        camera_position : tuple or list, optional
            (x, y, z) coordinates for the camera position.
        focal_point : tuple, list, or str, optional
            If tuple/list: (x, y, z) coordinates to look at.
            If str: The name of the surface in self.os to look at.
            If None: Defaults to center of system.
        view_up : tuple or list, optional
            (x, y, z) vector defining the "up" direction. Default is (0, 0, 1).
        orthographic : bool, optional
            If True, start with a parallel projection. 
            If False, start with a perspective projection. Default is False.
        """

        # Set opacity based on number of ray actors
        # N.B.: this overestimates the number of photons.
        # N.B.: _set_ray_opacity will not update _ray_opacity
        #       at each start render the opacity will be set automatically
        #       unless an opacity is specified with set_ray_opacity
        if self._ray_opacity is None:
            n_rays = 0
            for actor in self.actors:
                if actor.startswith('rays_'):
                    n_rays += self.actors[actor].GetMapper().GetDataSetInput().GetLines().GetNumberOfCells()
            opacity = self._get_sigmoid_opacity(n_rays)
            self._set_ray_opacity(opacity)

        # Rendering
        renderer = vtk.vtkRenderer()
        renderer.SetBackground(0.1, 0.15, 0.25)

        inf = float('inf')
        sys_bounds = [inf, -inf, inf, -inf, inf, -inf]
        has_system_actors = False

        for name, actor in self.actors.items():
            renderer.AddActor(actor)
            
            # Include this actor in the bounds calculation
            if not name.startswith("rays"):
                has_system_actors = True
                b = actor.GetBounds() # (xmin, xmax, ymin, ymax, zmin, zmax)
                sys_bounds[0] = min(sys_bounds[0], b[0])
                sys_bounds[1] = max(sys_bounds[1], b[1])
                sys_bounds[2] = min(sys_bounds[2], b[2])
                sys_bounds[3] = max(sys_bounds[3], b[3])
                sys_bounds[4] = min(sys_bounds[4], b[4])
                sys_bounds[5] = max(sys_bounds[5], b[5])

        # Display instructions
        desc = [
            "Press: 'q' to exit",
            "Press: 'r' to reset the camera zoom",
            "Press: 'o' to toggle orthographic/perspective view",
            "Press: 'l' to select/deselect actor under cursor",
            "Press: 'c' to move the focal point on selected actors",
            "Press: 'h' to hide the selected actors", 
            "Press: 'i' to isolate selected actors (and hide others)",
            "Press: 'Esc' to clear selection and actions",
            "Press: 'x', 'y' or 'z' to align the up-vector to the desired axis",
            "Press: 'w', 's' or 'p' to switch to wireframe, surface or points representation",
        ]
        textActor = vtk.vtkTextActor()
        textActor.SetInput('\n'.join(desc))
        position_coordinate = textActor.GetPositionCoordinate()
        position_coordinate.SetCoordinateSystemToNormalizedViewport()
        position_coordinate.SetValue(0.01, 0.99)
        textActor.GetTextProperty().SetJustificationToLeft()
        textActor.GetTextProperty().SetVerticalJustificationToTop()
        textActor.GetTextProperty().SetFontSize(18)
        textActor.GetTextProperty().SetColor(vtk.vtkNamedColors().GetColor3d("Gold"))
        textActor.GetTextProperty().SetFontFamily(vtk.VTK_COURIER)
        textActor.SetVisibility(False)
        renderer.AddActor2D(textActor)

        hintActor = vtk.vtkTextActor()
        hintActor.SetInput("Press '?' for instructions")
        hprop = hintActor.GetTextProperty()
        hprop.SetFontSize(16)
        hprop.SetColor(vtk.vtkNamedColors().GetColor3d("Gold"))
        hprop.SetFontFamily(vtk.VTK_COURIER)
        hprop.SetJustificationToLeft()
        hprop.SetVerticalJustificationToTop()
        hcoord = hintActor.GetPositionCoordinate()
        hcoord.SetCoordinateSystemToNormalizedViewport()
        hcoord.SetValue(0.01, 0.99)
        hintActor.SetVisibility(True)
        renderer.AddActor2D(hintActor)

        render_window = vtk.vtkRenderWindow()
        render_window.SetWindowName(self.os.name)
        render_window.AddRenderer(renderer)
        render_window.SetSize(*self.window_size)
        render_window_interactor = vtk.vtkRenderWindowInteractor()
        render_window_interactor.SetRenderWindow(render_window)

        # Camera
        cam_style = vtk.vtkInteractorStyleTrackballCamera()
        render_window_interactor.SetInteractorStyle(cam_style)

        # Axes
        axes = vtk.vtkAxesActor()
        widget = vtk.vtkOrientationMarkerWidget()
        rgba = [0] * 4
        vtk.vtkNamedColors().GetColor('Carrot', rgba)
        widget.SetOutlineColor(rgba[0], rgba[1], rgba[2])
        widget.SetOrientationMarker(axes)
        widget.SetInteractor(render_window_interactor)
        widget.SetEnabled(1)
        widget.InteractiveOn()

        selected_actors = {}
        is_isolated = False
        picker = vtk.vtkCellPicker()
        picker.SetTolerance(0.005)

        def perform_view_reset(up_vector):
            active_camera = renderer.GetActiveCamera()
            if not active_camera:
                return

            # Normalize up_vector
            norm_up = np.sqrt(sum(x*x for x in up_vector))
            if norm_up == 0: return
            target_up = [x/norm_up for x in up_vector]

            # Get current view plane normal
            vpn = active_camera.GetViewPlaneNormal()
            
            # Compute dot product
            dot = sum(vpn[i] * target_up[i] for i in range(3))
            
            # If dot product is near 1 or -1, vectors are parallel.
            if abs(dot) > 0.95:
                fp = active_camera.GetFocalPoint()
                dist = active_camera.GetDistance()
                
                # Move camera to a perpendicular axis based on target UP
                if abs(target_up[2]) > 0.9: 
                    active_camera.SetPosition(fp[0], fp[1] - dist, fp[2])
                elif abs(target_up[1]) > 0.9:
                    active_camera.SetPosition(fp[0] - dist, fp[1], fp[2])
                else:
                    active_camera.SetPosition(fp[0], fp[1], fp[2] + dist)

                if has_system_actors:
                    renderer.ResetCamera(sys_bounds)
                else:
                    renderer.ResetCamera()

            active_camera.SetViewUp(*up_vector)
            active_camera.OrthogonalizeViewUp() 
            render_window.Render()

        def key_press_callback(caller, event):
            interactor = caller
            key_sym = interactor.GetKeySym()
            
            if key_sym is None:
                return
            
            # Reset view ('r')
            if key_sym.lower() == 'r':
                if has_system_actors:
                    renderer.ResetCamera(sys_bounds)
                else:
                    renderer.ResetCamera()
                render_window.Render()
                interactor.SetKeySym("") 
                interactor.SetKeyCode("\0")
            
            # Instructions ('?')
            elif key_sym.lower() == 'question':
                current_vis = textActor.GetVisibility()
                # Toggle: If currently visible, make invisible.
                textActor.SetVisibility(not current_vis)
                # Hint is always the opposite of Instructions
                hintActor.SetVisibility(current_vis)
                render_window.Render()

            # Orthographic/perspective ('o')
            elif key_sym.lower() == 'o':
                cam = renderer.GetActiveCamera()
                if cam.GetParallelProjection():
                    # Switching Parallel -> Perspective
                    scale = cam.GetParallelScale()
                    fov = cam.GetViewAngle()
                    if fov > 0:
                        target_dist = scale / np.tan(np.deg2rad(fov) / 2.0)
                        fp = np.array(cam.GetFocalPoint())
                        pos = np.array(cam.GetPosition())
                        direction = fp - pos
                        dist_mag = np.linalg.norm(direction)
                        if dist_mag > 0:
                            direction_norm = direction / dist_mag
                            cam.SetPosition(fp - direction_norm * target_dist)
                    cam.SetParallelProjection(False)
                else:
                    # Switching Perspective -> Parallel
                    dist = cam.GetDistance()
                    fov = cam.GetViewAngle()
                    new_scale = dist * np.tan(np.deg2rad(fov) / 2.0)
                    cam.SetParallelScale(new_scale)
                    cam.SetParallelProjection(True)
                
                renderer.ResetCameraClippingRange()
                render_window.Render()
            
            # Selection ('l')
            elif key_sym.lower() == 'l':
                click_pos = interactor.GetEventPosition()
                picker.Pick(click_pos[0], click_pos[1], 0, renderer)
                picked_actor = picker.GetActor()
                if picked_actor:
                    if picked_actor in selected_actors:
                        # Deselect
                        original_color = selected_actors.pop(picked_actor)
                        picked_actor.GetProperty().SetColor(*original_color)
                    else:
                        # Select
                        current_color = picked_actor.GetProperty().GetColor()
                        selected_actors[picked_actor] = current_color
                        picked_actor.GetProperty().SetColor(0.0, 1.0, 0.0) # Green
                render_window.Render()
            
            # Center ('c')
            elif key_sym.lower() == 'c':
                if selected_actors:
                    # Calculate center of all selected actors
                    sel_bounds = [inf, -inf, inf, -inf, inf, -inf]
                    for actor in selected_actors:
                        b = actor.GetBounds()
                        sel_bounds[0] = min(sel_bounds[0], b[0])
                        sel_bounds[1] = max(sel_bounds[1], b[1])
                        sel_bounds[2] = min(sel_bounds[2], b[2])
                        sel_bounds[3] = max(sel_bounds[3], b[3])
                        sel_bounds[4] = min(sel_bounds[4], b[4])
                        sel_bounds[5] = max(sel_bounds[5], b[5])
                    
                    center_x = (sel_bounds[0] + sel_bounds[1]) / 2.0
                    center_y = (sel_bounds[2] + sel_bounds[3]) / 2.0
                    center_z = (sel_bounds[4] + sel_bounds[5]) / 2.0
                    
                    # Move Focal Point
                    renderer.GetActiveCamera().SetFocalPoint(center_x, center_y, center_z)
                    renderer.ResetCameraClippingRange()
                    render_window.Render()

            # Hide selected ('h')
            elif key_sym.lower() == 'h':
                # Create a reverse lookup to find surface names from actor objects
                actor_to_name = {v: k for k, v in self.actors.items()}

                for actor in list(selected_actors.keys()):
                    # Hide the surface
                    original_color = selected_actors.pop(actor)
                    actor.GetProperty().SetColor(*original_color)
                    actor.SetVisibility(False)
                    
                    # Find and hide associated hits
                    if actor in actor_to_name:
                        
                        surface_name = actor_to_name[actor]
                        
                        # Look for keys starting with "hits_" + surface_name
                        target_prefix = f"hits_{surface_name}"
                        
                        for key, hit_actor in self.actors.items():
                            if key == target_prefix or key.startswith(target_prefix + "_"):
                                hit_actor.SetVisibility(False)

                render_window.Render()

            # Isolate selected ('i')
            elif key_sym.lower() == 'i':
                nonlocal is_isolated
                
                # If we have a selection, Isolate it and then Deselect
                if selected_actors:
                    is_isolated = True
                    # Hide everything not in selection
                    for name, actor in self.actors.items():
                        # Always keep rays visible
                        if "rays_" in name:
                            actor.SetVisibility(True)
                            continue
                        
                        if actor in selected_actors:
                            actor.SetVisibility(True)
                            # Restore color immediately
                            orig_col = selected_actors[actor]
                            actor.GetProperty().SetColor(*orig_col)
                        else:
                            actor.SetVisibility(False)
                    
                    # Clear selection since we are now in "Isolation Mode" with no active targets
                    selected_actors.clear()
                    
                # If nothing selected, 'i' acts as "Show All" if currently isolated
                elif is_isolated:
                    is_isolated = False
                    for actor in self.actors.values():
                        actor.SetVisibility(True)
                
                render_window.Render()

                # Consume the key, otherwise the axes will disappear (IDKW)
                interactor.SetKeySym("") 
                interactor.SetKeyCode("\0")

            # Clear selections ('Escape')
            elif key_sym == 'Escape':
                is_isolated = False
                # Restore colors of selected
                for actor, color in selected_actors.items():
                    actor.GetProperty().SetColor(*color)
                selected_actors.clear()
                # Ensure everything is visible
                for actor in self.actors.values():
                    actor.SetVisibility(True)
                render_window.Render()

            # Axes and modes
            elif key_sym.lower() == 'y':
                perform_view_reset((0,1,0))
            elif key_sym.lower() == 'x':
                perform_view_reset((1,0,0))
            elif key_sym.lower() == 'z':
                perform_view_reset((0,0,1))
            elif key_sym.lower() == 'w':
                for actor in renderer.GetActors():
                    actor.GetProperty().SetRepresentationToWireframe()
                render_window.Render()
            elif key_sym.lower() == 's':
                for actor in renderer.GetActors():
                    actor.GetProperty().SetRepresentationToSurface()
                render_window.Render()
            elif key_sym.lower() == 'p':
                for actor in renderer.GetActors():
                    actor.GetProperty().SetRepresentationToPoints()
                render_window.Render()
            elif key_sym.lower() == 'q':
                    quit(render_window_interactor)

        priority = 1.0
        render_window_interactor.AddObserver(vtk.vtkCommand.CharEvent, key_press_callback, priority)

        # Start
        render_window.Render()
        
        # Camera configuration
        cam = renderer.GetActiveCamera()
        
        # Apply initial projection type
        cam.SetParallelProjection(orthographic)

        if camera_position is not None:
            cam.SetPosition(*camera_position)
            cam.SetViewUp(*view_up)
            
            if focal_point is not None:
                if isinstance(focal_point, str):
                    fp = self.os[focal_point].position
                    if self._tel_to_local_rot is not None:
                        fp = self._tel_to_local_rot @ fp + self._translation_vector
                    cam.SetFocalPoint(*fp)
                else:
                    cam.SetFocalPoint(*focal_point)
            elif has_system_actors:
                cx = (sys_bounds[0] + sys_bounds[1]) / 2.0
                cy = (sys_bounds[2] + sys_bounds[3]) / 2.0
                cz = (sys_bounds[4] + sys_bounds[5]) / 2.0
                cam.SetFocalPoint(cx, cy, cz)
            else:
                cam.SetFocalPoint(0, 0, 0)
            
            renderer.ResetCameraClippingRange()
        else:
            if has_system_actors:
                renderer.ResetCamera(sys_bounds)
            else:
                renderer.ResetCamera()
        
        # Increase far clip plane
        rng = cam.GetClippingRange()
        cam.SetClippingRange(rng[0], rng[1] * 1000) 

        render_window.Render()
        render_window_interactor.Initialize()
        render_window_interactor.Start()

        # Stop
        quit(render_window_interactor)

    def _create_wavelength_lut(self):
        """Creates a color transfer function for wavelength in the range 200nm - 1000nm."""
        ctf = vtk.vtkColorTransferFunction()
        ctf.SetColorSpaceToRGB()
        
        # Bright white/blue
        ctf.AddRGBPoint(200.0, 0.9, 0.9, 1.0) 
        # Electric purple
        ctf.AddRGBPoint(300.0, 0.7, 0.5, 1.0)   
        
        # Standard Rainbow, fully saturated
        ctf.AddRGBPoint(380.0, 0.5, 0.0, 1.0)   # Violet
        ctf.AddRGBPoint(440.0, 0.0, 0.2, 1.0)   # Blue
        ctf.AddRGBPoint(490.0, 0.0, 1.0, 1.0)   # Cyan
        ctf.AddRGBPoint(510.0, 0.2, 1.0, 0.2)   # Green
        ctf.AddRGBPoint(580.0, 1.0, 1.0, 0.0)   # Yellow
        ctf.AddRGBPoint(600.0, 1.0, 0.5, 0.0)   # Orange
        ctf.AddRGBPoint(700.0, 1.0, 0.0, 0.0)   # Red
        
        # Bright cherry red
        ctf.AddRGBPoint(850.0, 0.9, 0.2, 0.2)   
        # Pale pink
        ctf.AddRGBPoint(1000.0, 0.7, 0.5, 0.5)
        
        return ctf

    def add_rays(self, start, stop, surface_id=None, wavelengths=None, directions=None, length=None, point_size=1.0, show_rays=True, show_hits=True):
        """
        Draw rays from start to stop positions and highlight stop points.
        
        If a stop position is NaN (indicating a miss), the ray is skipped by default.
        If 'length' and 'directions' are provided, rays with NaN stops are drawn 
        starting from 'start' along 'direction' for 'length' units (without a hit dot).
        
        Parameters
        ----------
        start : ndarray
            (n, 3) array of starting coordinates (x, y, z).
        stop : ndarray
            (n, 3) array of stopping coordinates (x, y, z).
        surface_id: ndarray
            (n,) array of surface indices reached by each ray.
        directions : ndarray, optional
            (n, 3) array of direction vectors. Required if plotting NaN rays with fixed length.
        length : float, optional
            If provided, rays with NaN stop will be drawn with this length using the direction vector.
        opacity : float, optional
            Transparency of the rays from 0.0 (invisible) to 1.0 (opaque). Default is 0.5.
        point_size : float, optional
            Size of the hit points. Default is 1.0
        show_rays : bool, optional
            Whether to show rays or not. Default is True.
        show_hits : bool, optional
            Whether to show hits or not. Default is True.
        """

        if not show_hits and not show_rays:
            return

        has_directions = directions is not None

        has_wavelengths = wavelengths is not None

        if start.shape != stop.shape:
            raise ValueError("Start and stop points must have the same shape.")

        if has_directions and start.shape != directions.shape:
            raise ValueError("Start/stop poinnts and directions must have the same shape.")
        
        if has_wavelengths and start.shape[0] != wavelengths.shape[0]:
            print(start.shape[0], wavelengths.shape[0])
            raise ValueError("Start/stop points and wavelengths shape mismatch.")
        
        wavelength_scalars = vtk.vtkFloatArray()
        wavelength_scalars.SetName("Wavelengths")
        
        n_photons = start.shape[0]
        
        # Rays
        points_rays = vtk.vtkPoints()
        lines_rays = vtk.vtkCellArray()
        
        # Hits
        hits_collections = {}
        if surface_id is None:
            surface_id = np.zeros((n_photons,), dtype=np.int32)

        # Check which rays have intersected a surface
        valid_stops = ~np.any(np.isnan(stop), axis=1)
        
        for i in range(n_photons):
            current_start = start[i]
            current_stop = stop[i]
            label = self.os[surface_id[i]].name
            
            is_valid_stop = valid_stops[i]
            
            final_stop = None
            draw_hit = False
            if is_valid_stop:
                # Draw from start to stop and add a hit
                final_stop = current_stop
                draw_hit = True
            elif length is not None and has_directions:
                # Draw fixed length ray if directions are available
                final_stop = current_start + directions[i] * length
                draw_hit = False
            else:
                # Ignore not intersected ray
                continue
            
            if show_rays:
                ## Ray logic
                id_start = points_rays.InsertNextPoint(current_start)
                id_stop_line = points_rays.InsertNextPoint(final_stop)
                
                line = vtk.vtkLine()
                line.GetPointIds().SetId(0, id_start)
                line.GetPointIds().SetId(1, id_stop_line)
                lines_rays.InsertNextCell(line)

                # Store wavelength
                if has_wavelengths:
                    wavelength_scalars.InsertNextValue(wavelengths[i])
            
            ## Hit logic
            if draw_hit:
                if label not in hits_collections:
                    hits_collections[label] = {
                        'points': vtk.vtkPoints(),
                        'verts': vtk.vtkCellArray()
                    }
                # Add point to specific surface collection
                pt_id = hits_collections[label]['points'].InsertNextPoint(final_stop)
                hits_collections[label]['verts'].InsertNextCell(1)
                hits_collections[label]['verts'].InsertCellPoint(pt_id)
            
        ## Rays actor
        if show_rays:
            raysPolyData = vtk.vtkPolyData()
            raysPolyData.SetPoints(points_rays)
            raysPolyData.SetLines(lines_rays)
            
            mapper_rays = vtk.vtkPolyDataMapper()
            mapper_rays.SetInputData(raysPolyData)

            if has_wavelengths:
                raysPolyData.GetCellData().SetScalars(wavelength_scalars)                
                # Use the lookup table
                lut = self._create_wavelength_lut()
                mapper_rays.SetLookupTable(lut)
                mapper_rays.SetScalarRange(200, 1000) # nm
                mapper_rays.SetScalarModeToUseCellData()
            else:
                mapper_rays.ScalarVisibilityOff()
            
            actor_rays = vtk.vtkActor()
            actor_rays.SetMapper(mapper_rays)

            if not has_wavelengths:
                actor_rays.GetProperty().SetColor(0.0, 0.7, 1.0) # Electric blue
            
            actor_rays.GetProperty().SetLineWidth(1.0)
            if self._ray_opacity is not None:
                actor_rays.GetProperty().SetOpacity(float(self._ray_opacity))
            
            # Make rays unpickable
            actor_rays.PickableOff()

            base_key = 'rays'
            key = base_key
            count = 0
            while key in self.actors:
                key = f"{base_key}_{count}"
                count += 1
            
            self.actors[key] = actor_rays

        ## Hits actor
        if show_hits:
            for label, data in hits_collections.items():
                if data['points'].GetNumberOfPoints() == 0:
                    continue
            
                dotsPolyData = vtk.vtkPolyData()
                dotsPolyData.SetPoints(data['points'])
                dotsPolyData.SetVerts(data['verts'])
                
                mapper_dots = vtk.vtkPolyDataMapper()
                mapper_dots.SetInputData(dotsPolyData)
                
                actor_dots = vtk.vtkActor()
                actor_dots.SetMapper(mapper_dots)
                actor_dots.GetProperty().SetColor(1.0, 1.0, 0.0)
                actor_dots.GetProperty().SetPointSize(point_size)
                actor_dots.GetProperty().SetOpacity(1.0)
                actor_dots.PickableOff()

                # Naming convention: hits_SurfaceID
                # If multiple batches are added, append an index
                base_key = f"hits_{label}"
                key = base_key
                count = 0
                while key in self.actors:
                    key = f"{base_key}_{count}"
                    count += 1
                
                self.actors[key] = actor_dots

    def _set_ray_opacity(self, value):
        """Set the opacity of rays.
        
        Parameters
        ----------
        value: float
            Ray opacity.
        
        """
        if value is None:
            return
        
        for actor_name in self.actors:
            if actor_name.startswith("rays_"):
                self.actors[actor_name].GetProperty().SetOpacity(float(value))

    def set_ray_opacity(self, value):
        """Set the opacity of rays.
        
        Parameters
        ----------
        value: float
            Ray opacity.
        
        """
        self._ray_opacity = value
        self._set_ray_opacity(value)

    def _apply_global_transform(self):
        """
        Apply a global rotation and translation to all current actors.
        
        Parameters
        ----------
        R : ndarray
            (3, 3) Rotation matrix. 
        t : ndarray or list
            (3,) Translation vector (x, y, z).
        """
        # Convert numpy array to vtkMatrix4x4
        vtk_R = vtk.vtkMatrix4x4()
        for i in range(3):
            for j in range(3):
                vtk_R.SetElement(i, j, self._tel_to_local_rot[i, j])

        # Iterate over all actors
        for actor in self.actors.values():
            
            # Get the existing transform
            transform = actor.GetUserTransform()
            if transform is None:
                transform = vtk.vtkTransform()
                transform.SetMatrix(actor.GetMatrix())
                actor.SetUserTransform(transform)
            
            # Apply transformations
            # PostMultiply ensures we are applying this to the "Global" world coordinates
            transform.PostMultiply()
            
            # Apply rotation
            transform.Concatenate(vtk_R)
            
            # Apply translation
            transform.Translate(*self._translation_vector)

            transform.Update()


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
import numpy as np
import cupy as cp

coordinate_indices = {
    'X': 0, 'x': 0,
    'Y': 1, 'y': 1,
    'Z': 2, 'z': 2,
}

def scatter(vectors3d, colors_array=None, plane='xy', s=0.1, marker=None, ax=None, scale=1., cbar=True, **kwargs):
    """Creates a scatter plot of two components of a set of 3D vectors.

    Parameters
    ----------
    vectors3d : array_like
        A numpy.ndarray or list of 3D vectors. The shape should be (N, 3), where N is the number of vectors.
    colors_array : array_like, optional
        An numpy.ndarray of color values for each point. Can be a list of color names, a list of RGB/RGBA tuples, or a 1D array of numerical values to be mapped to colors using a colormap.
    plane : str
        String representation of the components to plot, e.g.: 'xy', 'zy', 'zx', etc.
    s : float, optional
        The marker size in points**2 (default: 0.1).
    marker : str, optional
        The marker style (default: None, which uses the default marker style from matplotlib). See `matplotlib.markers` for valid marker styles.
    ax : matplotlib.axes.Axes, optional
        The axes on which to plot. If None, the current axes are used (plt.gca()).
    scale : float, optional
        A scaling factor applied to the x and y coordinates (default: 1.).
    cbar : bool, optional
        Whether to add a colorbar (default: True). The colorbar will only be added if `colors_array` is also provided and is a list or NumPy array.
    **kwargs :
        Additional keyword arguments passed to `matplotlib.pyplot.scatter`. This allows for customization of the plot, such as setting the colormap ('cmap'), edgecolors ('edgecolors'), etc.

    Returns
    -------
    cax : matplotlib.axes.Axes or None
        The axes of the colorbar if a colorbar is created; otherwise, None.
    """
    if ax is None:
        ax = plt.gca()
    
    if not isinstance(plane, str) or len(plane)>2:
        raise(ValueError("`plane` argument must indicate in which plane perform the scatter plot (e.g. 'xy', 'Yx', 'Xz', etc...)."))
    index1 = coordinate_indices[plane[0]]
    index2 = coordinate_indices[plane[1]]

    if isinstance(vectors3d, cp.ndarray):
        vectors3d = vectors3d.get()

    x = vectors3d[:,index1]/scale
    y = vectors3d[:,index2]/scale

    sc = ax.scatter(x, y, s=s, marker=marker, c=colors_array, **kwargs)
    ax.set_aspect(1)
    cax = None
    if isinstance(colors_array, (list, np.ndarray)) and cbar:
        ax_divider = make_axes_locatable(ax)
        cax = ax_divider.append_axes("right", size="5%", pad=f"{3}%")
        plt.colorbar(sc, cax=cax)
        plt.sca(ax)

    return cax


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

# -- Path setup --------------------------------------------------------------
import pathlib
import sys
import re

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
try: # iactsim is installed
    import iactsim
except ImportError: # iactsim is run from its source checkout
    sys.path.insert(0, pathlib.Path(__file__).parents[2].resolve().as_posix())
    import iactsim

# -- Project information -----------------------------------------------------
project = 'iactsim'
copyright = '2024-, Davide Mollica'
author = 'Davide Mollica'
release = re.match(r'v?(\d+\.\d+\.\d+)', iactsim.__version__).group(1)
version = iactsim.__version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

# -- General configuration ---------------------------------------------------
# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'myst_nb',
    'IPython.sphinxext.ipython_console_highlighting',
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',  # parse numpy-style docstrings
    'sphinx.ext.autosummary',
    'sphinx_toolbox.code',  # code blocks
    'sphinx.ext.intersphinx', # links to external projects
    'autodocsumm',
    'sphinx_autodoc_typehints',
]

# Avoid calling representation methods of default values
autodoc_preserve_defaults = True

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True
napoleon_preprocess_types = False
napoleon_type_aliases = None
napoleon_attr_annotations = True

templates_path = ['_templates']
exclude_patterns = []

intersphinx_mapping = {
    'numpy': ('https://numpy.org/doc/stable/', None),
    'python': ('https://docs.python.org/3/', None),
    'matplotlib': ('https://matplotlib.org/stable/', None),
    'numba': ('https://numba.readthedocs.io/en/stable/', None),
    'cupy': ('https://docs.cupy.dev/en/stable/', None),
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"
# html_theme = "pydata_sphinx_theme"
html_static_path = ['_static']

def setup(app):
    app.add_css_file('rtd-overrides.css')